import numpy

from amuse.test import amusetest
from amuse.units import units, quantities
from amuse.datamodel.particles import Particles, Particle
# from amuse.community.sse.interface import SSE
from amuse.community.seba.interface import SeBa

from amuse.ext import stellar_wind

from amuse.support.console import set_preferred_units


from amuse.support.console import set_printing_strategy


class TestStellarWind(amusetest.TestCase):
    def setUp(self):
        set_preferred_units(units.MSun, units.RSun, units.yr, units.kms,
                            units.kms**2, units.m/units.s**2, units.W, units.erg)

    def skip_no_scipy(self):
        try:
            import scipy
        except ImportError:
            self.skip("Failed to import scipy")

    def assertAllAlmostEqual(self, values, expected, *args, **kwargs):
        for value in values:
            self.assertAlmostEqual(value, expected, *args, **kwargs)

    def assertDecreasing(self, array):
        delta = array[1:] - array[:-1]
        for d in delta:
            self.assertLess(d, quantities.zero)

    def create_star(self, N=1):
        star = Particles(N)
        star.mass = 2 | units.MSun
        star.radius = 2 | units.RSun
        star.temperature = 5000 | units.K
        star.position = [1, 1, 1] | units.parsec
        star.velocity = [-1000, 0, 1000] | units.ms
        star.wind_mass_loss_rate = 1e-6 | units.MSun / units.yr
        star.initial_wind_velocity = 50 | units.ms
        star.terminal_wind_velocity = 500 | units.ms
        return star

    def create_stars_without_wind_attributes(self):
        stars = Particles(2)
        stars.mass = 2 | units.MSun
        stars.radius = 2 | units.RSun
        stars.temperature = 5000 | units.K
        stars.luminosity = 2 | units.LSun
        stars.age = 0 | units.Myr
        stars[0].position = [1, 1, 1] | units.parsec
        stars[1].position = [-1, -1, -1] | units.parsec
        stars[0].velocity = [-1000, 0, 1000] | units.ms
        stars[1].velocity = [0, 0, 0] | units.ms

        return stars

    def tearDown(self):
        set_printing_strategy('default')


class TestStarsWithMassLoss(TestStellarWind):
    def test_add_particles(self):
        star_particles = self.create_star(2)
        print("created parts")
        stars = stellar_wind.StarsWithMassLoss()
        print("created stars")
        stars.add_particles(star_particles)
        print("added", stars)

        self.assertEqual(stars.lost_mass, [0, 0] | units.MSun)
        self.assertEqual(stars.wind_release_time, [0, 0] | units.yr)

        self.assertAlmostEqual(stars.mu[0], 1.145934546 | units.amu)

    def test_add_particle(self):
        star_particles = self.create_star(1)
        stars = stellar_wind.StarsWithMassLoss()
        star = stars.add_particle(star_particles[0])

        self.assertEqual(star.lost_mass, 0 | units.MSun)
        self.assertEqual(star.wind_release_time, 0 | units.yr)

        self.assertAlmostEqual(star.mu, 1.145934546 | units.amu)

        attrs = stars.get_attribute_names_defined_in_store()
        self.assertFalse("mechanical_energy" in attrs)
        self.assertFalse("previous_mechanical_luminosity" in attrs)

    def test_default_lmech(self):
        star_particles = self.create_star(1)
        stars = stellar_wind.StarsWithMassLoss()
        stars.track_mechanical_energy()
        star = stars.add_particle(star_particles[0])

        self.assertEqual(star.mechanical_energy, 0 | units.J)

    def test_add_particle_later(self):
        star_particles = self.create_star(2)
        stars = stellar_wind.StarsWithMassLoss()
        stars.add_particle(star_particles[0])

        stars.evolve_mass_loss(1 | units.yr)

        new_star = stars.add_particle(star_particles[1])
        self.assertAlmostEqual(new_star.wind_release_time, 1 | units.yr)

    def test_superfluous_attributes(self):
        star_particles = self.create_star(2)
        star_particles.other_attr = 10 | units.yr

        stars = stellar_wind.StarsWithMassLoss()
        stars.add_particles(star_particles)

        self.assertRaises(AttributeError, stars.__getattr__, "other_attr")

    def test_add_wrong_attribute(self):
        star_particles = self.create_star(2)
        star_particles

        stars = stellar_wind.StarsWithMassLoss()
        stars.add_particles(star_particles)

        self.assertRaises(AttributeError, stars.__setattr__, "other_attr", 10)


class FakeStevCode(object):
    def __init__(self, *args, **kwargs):
        self.particles = Particles()
        self.model_time = 0 | units.Myr
        self.mass_loss_rate = 1e-11 | units.MSun/units.yr

    def evolve_model(self, time):
        dt = time - self.model_time
        self.model_time = time
        self.particles.mass -= self.mass_loss_rate * dt
        self.particles.age = time


class TestPositionGenerator(TestStellarWind):
    def test_random_cube(self):
        numpy.random.seed(1232367)
        gen = stellar_wind.PositionGenerator('random')
        cube = gen.cube_generator(1000)
        self.assertEqual(cube.shape, (1000, 3))
        self.assertAlmostEqual(cube.min(), -0.99970440813)
        self.assertAlmostEqual(cube.max(), 0.997978975085)

    def test_regular_cube(self):
        numpy.random.seed(1232367)
        gen = stellar_wind.PositionGenerator(grid_type="regular")
        cube = gen.cube_generator(25)
        print(cube)
        self.assertEqual(cube.shape, (25, 3))
        self.assertAlmostEqual(cube.min(), -2./3.)
        self.assertAlmostEqual(cube.max(), 2./3.)

    def test_cutout_sphere(self):
        points = [[0., 0., 0.],
                  [1., 1., 1.],
                  [-1., -1., -1.],
                  [-0., 0., -.99],
                  [0., .5, 0.],
                  ]
        points = numpy.array(points)

        gen = stellar_wind.PositionGenerator()
        remaining = gen.cutout_sphere(points, 0.1)

        self.assertEqual(len(remaining), 2)
        self.assertEqual(remaining, points[-2:])

    def test_uniform_hollow_sphere(self):
        numpy.random.seed(1232367)
        gen = stellar_wind.PositionGenerator()
        points = gen.uniform_hollow_sphere(2, 0.6)
        self.assertEqual(len(points), 2)

        points = gen.uniform_hollow_sphere(1000, 0.1)
        self.assertEqual(len(points), 1000)

    def test_regular_hollow_sphere(self):
        numpy.random.seed(1232367)
        gen = stellar_wind.PositionGenerator(grid_type="regular")
        points = gen.uniform_hollow_sphere(2, 0.6)
        self.assertEqual(len(points), 2)

        points = gen.uniform_hollow_sphere(1000, 0.1)
        self.assertEqual(len(points), 1000)

    def test_density_distribution(self):
        numpy.random.seed(1234567)
        N = 100000
        rmin = 1 | units.RSun
        rmax = 10 | units.RSun
        gen = stellar_wind.PositionGenerator()
        p, _ = gen.generate_positions(N, rmin, rmax)
        r = p.lengths()

        print('rmin', r.min())
        print('rmax', r.max())
        print(r.mean())
        self.assertEqual(len(r), N)
        self.assertGreaterEqual(r.min(), rmin)
        self.assertLessEqual(r.max(), rmax)
        self.assertAlmostEqual(r.mean(), 5.49955427602 | units.RSun)

        return

        r = r.value_in(units.RSun)
        n_bins = 50
        n, bins = numpy.histogram(r, n_bins, range=(1, 10))
        bin_volume = 4./3. * numpy.pi * (bins[1:]**3 - bins[:-1]**3)
        dens = n / bin_volume
        bin_means = (bins[1:] + bins[:-1])/2.

        s = p[abs(p[:, 2]) < (0.1 | units.RSun)]
        from matplotlib import pyplot
        from amuse import plot as aplot
        aplot.plot(s[:, 0], s[:, 1], '.')
        pyplot.axis('equal')
        pyplot.savefig("scatter.pdf")
        pyplot.clf()

        x = numpy.linspace(1, 10, num=200)
        y = 300. / x**2
        # y = 0. * x + dens.mean()

        from matplotlib import pyplot
        # pyplot.plot(x, y)
        pyplot.loglog(x, y)
        pyplot.plot(bin_means, dens, '*')
        pyplot.savefig("dens.pdf")

    def test_alternate_radius_function(self):
        numpy.random.seed(123)
        gen = stellar_wind.PositionGenerator()
        # func = stellar_wind.LogisticVelocityAcceleration().radius_from_number
        func = stellar_wind.ConstantVelocityAcceleration().radius_from_number

        N = 100000
        rmin = 2 | units.RSun
        rmax = 6 | units.RSun
        star = Particle()
        star.radius = rmin
        star.acc_cutoff = 10 | units.RSun
        star.initial_wind_velocity = 4 | units.kms
        star.terminal_wind_velocity = 12 | units.kms

        p, _ = gen.generate_positions(N, rmin, rmax, func, star)
        r = p.lengths()

        self.assertEqual(len(r), N)
        self.assertGreaterEqual(r.min(), rmin)
        self.assertLessEqual(r.max(), rmax)
        print(r.mean())
        self.assertAlmostEqual(r.mean(), 4.00196447056 | units.RSun)

        return

        r = r.value_in(units.RSun)
        n_bins = 50
        n, bins = numpy.histogram(r, n_bins)
        bin_volume = 4./3. * numpy.pi * (bins[1:]**3 - bins[:-1]**3)
        dens = n / bin_volume
        bin_means = (bins[1:] + bins[:-1])/2.

        from matplotlib import pyplot
        pyplot.plot(bin_means, dens, '*')
        pyplot.savefig("dens.pdf")

    def test_random_rotation(self):
        numpy.random.seed(1232367)
        gen = stellar_wind.PositionGenerator()
        axis, angle = gen.random_rotation()
        print(axis)
        print(angle)
        self.assertEqual(len(axis), 3)
        self.assertAlmostEqual((axis**2).sum(), 1.)
        self.assertGreaterEqual(angle, 0.)
        self.assertLessEqual(angle, 2. * numpy.pi)

    def test_rotation_matrix(self):
        gen = stellar_wind.PositionGenerator()
        axis = [0, 0, 1]
        angle = numpy.pi
        matrix = gen.rotation_matrix(axis, angle)
        print(numpy.rint(matrix))

        expected = [[-1, 0, 0], [0, -1, 0], [0, 0, 1]]
        self.assertAlmostEqual(matrix, expected)

    def test_rotate_positions(self):
        gen = stellar_wind.PositionGenerator()
        axis = [0, 1, 0]
        angle = numpy.pi
        positions = [[0.5, 0.5, 0.5], [0.2, -0.5, 1.3]]
        rotated = gen.rotate_positions(positions, axis, angle)
        print(rotated)
        expected = [[-0.5, 0.5, -0.5], [-0.2, -0.5, -1.3]]
        self.assertAlmostEqual(rotated, expected)


class TestSimpleWind(TestStellarWind):
    def test_wind_particles(self):
        star = self.create_star()
        star_wind = stellar_wind.new_stellar_wind(1e-8 | units.MSun)
        star_wind.particles.add_particles(star)

        star_wind.evolve_model(1 | units.yr)
        wind = star_wind.create_wind_particles()

        self.assertEqual(len(wind), 100)
        min_dist = (wind.position - star.position).lengths().min()
        max_dist = (wind.position - star.position).lengths().max()
        self.assertGreaterEqual(min_dist, 2 | units.RSun)
        self.assertLessEqual(max_dist, 24.7 | units.RSun)
        plus_x_wind = wind[wind.x > star.x]
        minus_x_wind = wind[wind.x < star.x]
        plus_y_wind = wind[wind.y > star.y]
        minus_y_wind = wind[wind.y < star.y]
        plus_z_wind = wind[wind.z > star.z]
        minus_z_wind = wind[wind.z < star.z]
        self.assertGreaterEqual(minus_x_wind.vx.min(), -1500 | units.ms)
        self.assertLessEqual(minus_x_wind.vx.max(), -1000 | units.ms)
        self.assertGreaterEqual(plus_x_wind.vx.min(), -1000 | units.ms)
        self.assertLessEqual(plus_x_wind.vx.max(), -500 | units.ms)
        self.assertGreaterEqual(minus_y_wind.vy.min(), -500 | units.ms)
        self.assertLessEqual(minus_y_wind.vy.max(), 0 | units.ms)
        self.assertGreaterEqual(plus_y_wind.vy.min(), 0 | units.ms)
        self.assertLessEqual(plus_y_wind.vy.max(), 500 | units.ms)
        self.assertGreaterEqual(minus_z_wind.vz.min(), 500 | units.ms)
        self.assertLessEqual(minus_z_wind.vz.max(), 1000 | units.ms)
        self.assertGreaterEqual(plus_z_wind.vz.min(), 1000 | units.ms)
        self.assertLessEqual(plus_z_wind.vz.max(), 1500 | units.ms)

    def test_create_wind_short_time(self):
        numpy.random.seed(123457)
        star = self.create_star()
        star_wind = stellar_wind.new_stellar_wind(
            1e-9 | units.MSun)
        star_wind.particles.add_particles(star)

        star_wind.evolve_model(0.01 | units.yr)
        wind = star_wind.create_wind_particles()

        self.assertEqual(len(wind), 10)
        radii = (wind.position - star.position).lengths()
        self.assertGreaterEqual(radii.min(), star.radius)

    def test_target_gas(self):
        star = self.create_star()
        gas = Particles()
        star_wind = stellar_wind.new_stellar_wind(
            1e-8 | units.MSun, target_gas=gas, timestep=1 | units.day)
        star_wind.particles.add_particles(star)

        star_wind.evolve_model(1 | units.yr)
        self.assertEqual(len(gas), 99)

        star_wind.evolve_model(1.5 | units.yr)
        self.assertEqual(len(gas), 149)

    def test_compensate_gravity(self):
        numpy.random.seed(123457)
        star = self.create_star()
        star_wind = stellar_wind.new_stellar_wind(
            1e-8 | units.MSun, compensate_gravity=True,
            grid_type="random", rotate=False)
        star_wind.particles.add_particles(star)

        star_wind.evolve_model(1 | units.yr)
        wind = star_wind.create_wind_particles()

        self.assertEqual(len(wind), 100)
        radii = (wind.position - star.position).lengths()
        wind_velocities = (wind.velocity - star.velocity).lengths()
        r_sort = radii.argsort()
        v_sorted = wind_velocities[r_sort]

        print(v_sorted)
        self.assertDecreasing(v_sorted)

        self.assertAlmostEqual(v_sorted[0], 615.945250201 | units.kms)
        self.assertAlmostEqual(v_sorted[-1], 176.050310315 | units.kms)

        return

        from matplotlib import pyplot
        from amuse import plot as aplot
        aplot.plot(radii[r_sort], v_sorted)
        pyplot.show()

    def test_later_star_add(self):
        star1 = self.create_star()
        gas = Particles()
        star_wind = stellar_wind.new_stellar_wind(
            1e-8 | units.MSun, target_gas=gas, timestep=1 | units.day)
        star_wind.particles.add_particles(star1)

        star_wind.evolve_model(1 | units.yr)
        self.assertEqual(len(gas), 99)

        star2 = self.create_star()[0]
        star2.x = 50 | units.parsec
        star_wind.particles.add_particle(star2)

        star_wind.evolve_model(1.5 | units.yr)
        self.assertEqual(len(gas), 198)
        star1_gas = gas[gas.x < 25 | units.parsec]
        star2_gas = gas - star1_gas
        self.assertEqual(len(star1_gas), 149)
        self.assertEqual(len(star2_gas), 49)

    def test_derive_from_evolution(self):
        stars = self.create_stars_without_wind_attributes()
        stars[1].temperature = 16000 | units.K
        stars[1].luminosity = 20 | units.LSun

        star_wind = stellar_wind.new_stellar_wind(
            1e-8 | units.MSun, derive_from_evolution=True, tag_gas_source=True)

        stars = star_wind.particles.add_particles(stars)
        stev = FakeStevCode()
        stev.mass_loss_rate = [1e-7, 2e-7] | units.MSun/units.yr
        stev.particles.add_particles(stars)
        chan = stev.particles.new_channel_to(
            star_wind.particles,
            attributes=["age", "radius", "mass", "luminosity", "temperature"])

        time = 0 | units.yr
        while time < 2.1 | units.yr:
            stev.evolve_model(time)
            chan.copy()
            star_wind.evolve_model(time)
            time += 0.5 | units.yr

        self.assertTrue(star_wind.has_new_wind_particles(),
                        "No new wind particles")
        wind = star_wind.create_wind_particles()
        self.assertEqual(len(wind), 59)

        wind_0 = wind[wind.source == stars[0].key]
        wind_1 = wind[wind.source == stars[1].key]
        self.assertEqual(len(wind_0), 19)
        self.assertEqual(len(wind_1), 40)

        vel_0 = (wind_0.velocity - stars[0].velocity).lengths()
        vel_1 = (wind_1.velocity - stars[1].velocity).lengths()
        self.assertAllAlmostEqual(vel_0, 617.83472287 | units.kms)
        self.assertAllAlmostEqual(vel_1, 864.87956297 | units.kms)

    def test_static_wind_from_evolution(self):
        stars = self.create_stars_without_wind_attributes()
        star_wind = stellar_wind.new_stellar_wind(
            1e-8 | units.MSun, derive_from_evolution=True)
        stev = FakeStevCode()
        stev.mass_loss_rate = [1e-11, 2e-11] | units.MSun/units.yr
        stev.particles.add_particles(stars)

        stellar_wind.static_wind_from_stellar_evolution(
            star_wind, stev, 1.49 | units.Gyr, 1.495 | units.Gyr)

        self.assertAlmostRelativeEquals(
            star_wind.particles.wind_mass_loss_rate,
            [1e-11, 2e-11] | units.MSun/units.yr,
            places=5
            )

        self.assertAlmostRelativeEquals(
            star_wind.particles.terminal_wind_velocity,
            [615.521, 613.198] | units.kms, places=5)


class ConstantSubclass(stellar_wind.AccelerationFunction):
    def acceleration_from_radius(self, radius, star):
        return numpy.zeros_like(radius, dtype=float) | units.m/units.s**2


class RSquaredSubclass(stellar_wind.AccelerationFunction):
    def scaling(self, star):
        return 0.5 * ((star.terminal_wind_velocity**2
                       - star.initial_wind_velocity**2)
                      / (1./star.radius - 1./star.acc_cutoff))

    def acceleration_from_radius(self, r, star):
        return self.scaling(star)/r**2


class TestAccelerationFunctions(TestStellarWind):
    def create_star(self):
        star = Particle()
        star.radius = 2 | units.RSun
        star.acc_cutoff = 10 | units.RSun
        star.initial_wind_velocity = 4 | units.kms
        star.terminal_wind_velocity = 12 | units.kms
        return star

    def test_default_acceleration_1(self):
        self.skip_no_scipy()

        func = ConstantSubclass()
        self.constant_asserts(func)

    def slowtest_default_acceleration_2(self):
        self.skip_no_scipy()

        func = RSquaredSubclass()
        self.r_squared_asserts(func)

    def test_constant_velocity(self):
        func = stellar_wind.ConstantVelocityAcceleration()
        self.constant_asserts(func)

    def test_r_squared_velocity(self):
        self.skip_no_scipy()

        func = stellar_wind.RSquaredAcceleration()
        self.r_squared_asserts(func)

    def constant_asserts(self, func):
        star = self.create_star()
        star.initial_wind_velocity = star.terminal_wind_velocity
        radii = [2., 4.] | units.RSun

        accelerations = func.acceleration_from_radius(radii, star)
        self.assertAlmostEqual(accelerations, [0., 0.] | units.m/units.s**2)

        velocities = func.velocity_from_radius(radii, star)
        self.assertAlmostEqual(velocities, [12., 12.] | units.kms)

        times = (numpy.array([0, 1, 2, 3]) * star.radius
                 / star.terminal_wind_velocity)
        new_radii = func.radius_from_time(times, star)
        self.assertAlmostEqual(new_radii, [2., 4., 6., 8.] | units.RSun)

        random_numbers = numpy.linspace(0., 1., 4)

        radii = func.radius_from_number(random_numbers, 5 | units.RSun, star)
        self.assertAlmostEqual(radii, [2, 3., 4., 5] | units.RSun)

    def r_squared_asserts(self, func):
        star = self.create_star()
        radii = [2., 4., 8., 10.] | units.RSun

        accelerations = func.acceleration_from_radius(radii, star)
        expected = (numpy.array([40., 10., 2.5, 1.6]) * 0.0014378145
                    | units.m/units.s**2)
        self.assertAlmostEqual(accelerations, expected)

        velocities = func.velocity_from_radius(radii, star)
        expected = [4., 9.79795897113, 11.6619037897, 12.] | units.kms
        self.assertAlmostEqual(velocities, expected)

        times = (numpy.array([0., 1., 2., 3.]) * star.radius
                 / star.initial_wind_velocity)
        new_radii = func.radius_from_time(times, star)
        exp = [2.0, 6.50863038599, 12.4140692531, 18.4140691706] | units.RSun
        self.assertAlmostEqual(new_radii, exp)

        x = numpy.linspace(0., 1., 4)
        new_radii = func.radius_from_number(x, 5 | units.RSun, star)
        expected = [2,  2.72665604616,  3.7783430146, 5] | units.RSun
        self.assertAlmostEqual(new_radii, expected)

    def test_radius_from_time(self):
        self.skip_no_scipy()

        func = stellar_wind.RSquaredAcceleration()
        star = Particle()
        star.radius = 312. | units.RSun
        star.acc_cutoff = 312*5. | units.RSun
        star.initial_wind_velocity = 2 | units.kms
        star.terminal_wind_velocity = 20 | units.kms

        times = [0., 25., 100.] | units.yr
        print("times", times)
        radii = func.radius_from_time(times, star)
        print("radii", radii)

        self.assertAlmostEqual(radii[0], 312. | units.RSun)
        self.assertAlmostEqual(radii[1], 22644.6086263 | units.RSun)
        self.assertAlmostEqual(radii[2], 90704.1183516 | units.RSun)

        return

        times = numpy.linspace(0., 1, 100) | units.yr
        radii = func.radius_from_time(times, star)

        print(radii)
        from matplotlib import pyplot
        from amuse import plot as aplot
        aplot.plot(times, radii/star.radius)
        pyplot.show()

    def test_radius_from_number(self):
        self.skip_no_scipy()

        star = Particle(
            initial_wind_velocity=200. | units.kms, mass=20. | units.MSun,
            radius=10. | units.RSun, terminal_wind_velocity=1000. | units.kms)
        star.acc_cutoff = 5. * star.radius
        r_max = 24.2642526438 | units.RSun
        x = 0.1

        func = stellar_wind.RSquaredAcceleration()
        radius = func.radius_from_number(x, r_max, star)
        self.assertAlmostEqual(radius,  10.6029030808 | units.RSun)


class TestAcceleratingWind(TestStellarWind):
    def test_wind_creation_constant(self):
        numpy.random.seed(123457)
        star = self.create_star()
        star.initial_wind_velocity = star.terminal_wind_velocity
        star_wind = stellar_wind.new_stellar_wind(
            3e-11 | units.MSun,
            mode="accelerate",
            acceleration_function="constant_velocity",
            )
        star_wind.particles.add_particles(star)

        star_wind.evolve_model(10 | units.day)
        wind = star_wind.create_wind_particles()

        self.assertEqual(len(wind), 912)
        radii = (wind.position - star.position).lengths()
        velocities = (wind.velocity - star.velocity).lengths()

        self.assertGreaterEqual(radii.min(), star.radius)
        self.assertLessEqual(radii.max(), 8.212 | units.RSun)
        self.assertAlmostEqual(radii.mean(), 2.32897955194 | units.RSun)

        self.assertAlmostEqual(velocities.min(), star.initial_wind_velocity)
        self.assertAlmostEqual(velocities.max(), star.initial_wind_velocity)

        return

        from matplotlib import pyplot
        from amuse import plot as aplot
        import pylab
        n, bins, patches = pylab.hist(radii.value_in(units.RSun), 10)
        pylab.show()
        pyplot.clf()
        aplot.scatter(radii, velocities)
        pyplot.show()

    def test_wind_creation_rsquared(self):
        self.skip_no_scipy()

        numpy.random.seed(123457)
        star = self.create_star()
        star_wind = stellar_wind.new_stellar_wind(
            3e-9 | units.MSun,
            mode="accelerate",
            acceleration_function="rsquared",
            )
        star_wind.particles.add_particles(star)

        star_wind.evolve_model(10 | units.day)
        wind = star_wind.create_wind_particles()

        self.assertEqual(len(wind), 9)
        radii = (wind.position - star.position).lengths()
        velocities = (wind.velocity - star.velocity).lengths()

        self.assertGreaterEqual(radii.min(), star.radius)
        self.assertLessEqual(radii.max(), 12.023 | units.RSun)
        self.assertAlmostEqual(radii.mean(), 2.06872960025 | units.RSun)

        self.assertGreaterEqual(velocities.min(), star.initial_wind_velocity)
        self.assertLessEqual(velocities.max(), star.terminal_wind_velocity)
        self.assertAlmostEqual(velocities.mean(), 0.103550548126 | units.kms)

    def test_acceleration(self):
        star = self.create_star()
        star.position = [1, 1, 1] | units.RSun
        star_wind = stellar_wind.new_stellar_wind(
            3e-11 | units.MSun,
            mode="accelerate",
            acceleration_function="rsquared",
            compensate_gravity=False,
            r_out_ratio=5,
            )
        star = star_wind.particles.add_particles(star)

        distances = [2, 4, 5, 10] | units.RSun
        accelerations = star_wind.acceleration(star, distances)
        expected = ([0.00011120596693, 2.78014917326e-5, 1.77929547088e-5, 0.]
                    | units.m/units.s**2)

        self.assertAlmostEqual(accelerations, expected)

    def test_gravity_compensation(self):
        star = self.create_star()
        star.position = [1, 1, 1] | units.RSun
        star_wind = stellar_wind.new_stellar_wind(
            3e-11 | units.MSun,
            mode="accelerate",
            acceleration_function="constant_velocity",
            compensate_gravity=True,
            )
        star = star_wind.particles.add_particles(star)

        distances = [2, 4, 5, 10, 20] | units.RSun
        accelerations = star_wind.acceleration(star, distances)
        expected = ([137.213699216, 34.3034248039, 21.9541918745,
                     5.48854796862, 1.37213699216] | units.m/units.s**2)

        self.assertAlmostEqual(accelerations, expected)

    def test_pressure_compensation(self):
        star = self.create_star()
        star.position = [1, 1, 1] | units.RSun
        star_wind = stellar_wind.new_stellar_wind(
            3e-11 | units.MSun,
            mode="accelerate",
            acceleration_function="constant_velocity",
            compensate_gravity=False,
            compensate_pressure=True
            )
        star = star_wind.particles.add_particles(star)

        distances = [2, 4, 5, 10] | units.RSun
        accelerations = star_wind.acceleration(star, distances)
        print(accelerations)
        expected = ([-0.0187296577097, -0.00371643479392, -0.00220802076676,
                     -0.000438126811] | units.m/units.s**2)

        self.assertAlmostEqual(accelerations, expected)

    def test_get_gravity_at_point(self):
        star = self.create_star()
        star.position = [1, 1, 1] | units.RSun
        star_wind = stellar_wind.new_stellar_wind(
            3e-11 | units.MSun,
            mode="accelerate",
            acceleration_function="rsquared",
            compensate_gravity=False,
            compensate_pressure=False,
            r_out_ratio=5,
            )
        star_wind.particles.add_particles(star)

        # unaffected points
        points = [[1, 1, -10], [1, 12, 0], [1, -10, 1],
                  [12, 0, 1]] | units.RSun
        x, y, z = points.transpose()
        ax, ay, az = star_wind.get_gravity_at_point(1, x, y, z)
        zeros = [0, 0, 0, 0] | units.m/units.s**2
        self.assertEqual(ax, zeros)
        self.assertEqual(ay, zeros)
        self.assertEqual(az, zeros)

        # affected points
        points = [[5.1, 1, 1], [1, 1, 5.1], [1, 7, 1], [1, 7, 8],
                  [-8, 1, 1]] | units.RSun
        x, y, z = points.transpose()
        ax, ay, az = star_wind.get_gravity_at_point(1, x, y, z)

        result = quantities.as_vector_quantity([ax, ay, az]).transpose()
        expected = [[2.64618600667e-05, 0.0, 0.0],
                    [0.0, 0.0, 2.64618600667e-05],
                    [0.0, 1.23562185478e-05, 0.0],
                    [0.0, 3.40573571553e-06, 3.97335833479e-06],
                    [-5.49165268791e-06, 0.0, 0.0]] | units.m / units.s**2
        self.assertAlmostEqual(result, expected)


class TestHeatingWind(TestStellarWind):
    def test_wind_creation(self):
        numpy.random.seed(123456789)
        star = self.create_star()
        feedback_efficiency = 0.01
        r_max = 10 | units.RSun
        star_wind = stellar_wind.new_stellar_wind(
            3e-10 | units.MSun,
            mode="heating",
            r_max=r_max,
            feedback_efficiency=feedback_efficiency,
            )
        star_wind.particles.add_particles(star)

        evolve_time = 1 | units.day
        expected_u = 0.5*feedback_efficiency*star.terminal_wind_velocity**2

        star_wind.evolve_model(evolve_time)
        wind = star_wind.create_wind_particles()

        self.assertEqual(len(wind), 9)
        max_dist = (wind.position - star.position).lengths().max()
        self.assertLessEqual(max_dist, r_max + star.radius)
        self.assertAlmostEqual(max_dist, 8.69397162579 | units.RSun)

        self.assertAllAlmostEqual(wind.u, expected_u)

        for t in numpy.arange(2, 10) * evolve_time:
            star_wind.evolve_model(t)
        wind = star_wind.create_wind_particles()

        self.assertEqual(len(wind), 73)
        max_dist = (wind.position - star.position).lengths().max()
        self.assertLessEqual(max_dist, r_max + star.radius)

        self.assertAllAlmostEqual(wind.u, expected_u)

    def setup_supernova(self):
        numpy.random.seed(123456789)

        stars = Particles(2)
        stars.mass = [9, 10] | units.MSun
        stars[0].position = [1, 1, 1] | units.parsec
        stars[1].position = [-1, -1, -1] | units.parsec
        stars[0].velocity = [-1000, 0, 1000] | units.ms
        stars[1].velocity = [0, 0, 0] | units.ms

        stev = SeBa()
        stars = stev.particles.add_particles(stars)

        r_max = .1 | units.parsec
        star_wind = stellar_wind.new_stellar_wind(
            3e-5 | units.MSun,
            mode="heating",
            r_max=r_max,
            derive_from_evolution=True,
            tag_gas_source=True
            )
        star_wind.particles.add_particles(stars)

        return stev, star_wind, stars

    def test_supernova(self):
        stev, star_wind, stars = self.setup_supernova()
        chan = stev.particles.new_channel_to(
            star_wind.particles,
            attributes=["age", "radius", "mass", "luminosity", "temperature", "stellar_type"])

        dt = 5 | units.Myr
        t = 0 | units.Myr
        t_end = 31 | units.Myr

        wind_N = []
        while t < t_end:
            stev.evolve_model(t)
            chan.copy()
            star_wind.evolve_model(t)
            if star_wind.has_new_wind_particles():
                wind = star_wind.create_wind_particles()
                wind_1 = wind[wind.source == stars[0].key]
                wind_2 = wind - wind_1
                wind_N.append([len(wind_1), len(wind_2)])
                if len(wind_2) > 0:
                    print("time", t, "wind energy", (wind_2.u * wind_2.mass).sum())
            else:
                wind_N.append([0, 0])

            t += dt

        self.assertEqual(wind_N, [[0, 0], [32, 45], [57, 59], [114, 130], [302, 635], [1231, 5810], [2981, 285777]])
        supernova = wind_2
        sn_energy = (supernova.u * supernova.mass).sum()
        self.assertAlmostRelativeEqual(sn_energy, 1e49 | units.erg, 2)

    def test_supernova_manual(self):
        stev, star_wind, stars = self.setup_supernova()
        stev.stopping_conditions.supernova_detection.enable()

        chan = stev.particles.new_channel_to(
            star_wind.particles,
            attributes=["age", "radius", "mass", "luminosity", "temperature"])

        dt = 5 | units.Myr
        t = 0 | units.Myr
        t_end = 31 | units.Myr

        wind_N = []
        wind_E = [] | units.erg
        while stev.model_time < t_end:
            stev.evolve_model(t)
            chan.copy()

            if stev.stopping_conditions.supernova_detection.is_set():
                key = stev.stopping_conditions.supernova_detection.particles(0)[0].key
                star = star_wind.particles._get_particle(key)
                star.mass_loss_type = "supernova"
                print("supernova detected at time", t, stev.model_time)

            star_wind.evolve_model(t)

            if star_wind.has_new_wind_particles():
                wind = star_wind.create_wind_particles()
                wind_2 = wind[wind.source == stars[1].key]
                wind_N.append(len(wind_2))
                wind_E.append((wind_2.u * wind_2.mass).sum())
                print(len(wind), len(wind_2))

            if not stev.stopping_conditions.supernova_detection.is_set():
                t += dt

        print(wind_N)
        print(wind_E)
        self.assertEqual(wind_N[-2:], [606194, 0])
        self.assertAlmostRelativeEqual(wind_E[-2:], [1.e49, 0] | units.erg, 7)



from amuse.support.interface import InCodeComponentImplementation

from amuse.test.amusetest import TestWithMPI
from amuse.test.amusetest import TestCase


import numpy
import sys
import os
import time
import pickle
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.rfi import python_code
from amuse.rfi.core import (
    PythonCodeInterface,
    LegacyFunctionSpecification,
    legacy_function,
)
from amuse.rfi.async_request import AsyncRequestsPool
from amuse.rfi.async_request import ASyncRequestSequence
from amuse.rfi.tools.create_python_worker import CreateAPythonWorker

from amuse.support import exceptions


class ForTestingInterface(PythonCodeInterface):

    def __init__(self, **options):
        PythonCodeInterface.__init__(self, implementation_factory=ForTestingImplementation, **options)

    @legacy_function
    def get_mass():
        function = LegacyFunctionSpecification()
        function.addParameter('index_of_the_particle', dtype='int32', direction=function.IN)
        function.addParameter('mass', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        function.id = 10
        return function

    @legacy_function
    def set_mass():
        function = LegacyFunctionSpecification()
        function.addParameter('index_of_the_particle', dtype='int32', direction=function.IN)
        function.addParameter('mass', dtype='float64', direction=function.IN, description="The new mass of the particle")
        function.result_type = 'int32'
        function.can_handle_array = True
        function.id = 11
        return function

    @legacy_function
    def new_particle():
        function = LegacyFunctionSpecification()
        function.addParameter('mass', dtype='float64', direction=function.IN, description="The new mass of the particle")
        function.addParameter('other', dtype='int32', direction=function.IN)
        function.addParameter('index_of_the_particle', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def delete_particle():
        function = LegacyFunctionSpecification()
        function.addParameter('index_of_the_particle', dtype='int32', direction=function.IN)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def echo_int():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in', dtype='int32', direction=function.IN)
        function.addParameter('int_out', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        function.id = 12
        return function

    @legacy_function
    def echo_double():
        function = LegacyFunctionSpecification()
        function.addParameter('double_in', dtype='float64', direction=function.IN)
        function.addParameter('double_out', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        function.id = 13
        return function

    @legacy_function
    def echo_string():
        function = LegacyFunctionSpecification()
        function.addParameter('string_in', dtype='string', direction=function.IN)
        function.addParameter('string_out', dtype='string', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        function.id = 14
        return function

    @legacy_function
    def echo_strings():
        function = LegacyFunctionSpecification()
        function.addParameter('string_inout1', dtype='string', direction=function.INOUT)
        function.addParameter('string_inout2', dtype='string', direction=function.INOUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        function.id = 15
        return function

    @legacy_function
    def echo_bool():
        function = LegacyFunctionSpecification()
        function.addParameter('string_in', dtype='bool', direction=function.IN)
        function.addParameter('string_out', dtype='bool', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        function.id = 16
        return function

    @legacy_function
    def sleep():
        function = LegacyFunctionSpecification()
        function.addParameter('number_of_seconds', dtype='float64', direction=function.IN)
        function.result_type = 'int32'
        function.can_handle_array = False
        return function

    @legacy_function
    def sum_doubles():
        function = LegacyFunctionSpecification()
        function.addParameter('double_in1', dtype='float64', direction=function.IN)
        function.addParameter('double_in2', dtype='float64', direction=function.IN, default=1.0)
        function.addParameter('double_out', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def multiply_ints():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in1', dtype='int32', direction=function.IN)
        function.addParameter('int_in2', dtype='int32', direction=function.IN)
        function.addParameter('int_out', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def print_string():
        function = LegacyFunctionSpecification()
        function.addParameter('string_in', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def print_error_string():
        function = LegacyFunctionSpecification()
        function.addParameter('string_in', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def get_position():
        function = LegacyFunctionSpecification()
        function.addParameter('index_of_the_particle', dtype='int32', direction=function.IN)
        function.addParameter('x', dtype='float64', direction=function.OUT)
        function.addParameter('y', dtype='float64', direction=function.OUT)
        function.addParameter('z', dtype='float64', direction=function.OUT)
        function.addParameter('length', 'int32', function.LENGTH)
        function.result_type = 'int32'
        function.must_handle_array = True
        return function

    @legacy_function
    def set_position():
        function = LegacyFunctionSpecification()
        function.addParameter('index_of_the_particle', dtype='int32', direction=function.IN)
        function.addParameter('x', dtype='float64', direction=function.IN)
        function.addParameter('y', dtype='float64', direction=function.IN)
        function.addParameter('z', dtype='float64', direction=function.IN)
        function.addParameter('length', 'int32', function.LENGTH)
        function.result_type = 'int32'
        function.must_handle_array = True
        return function

    @legacy_function
    def copy_over_interface():
        function = LegacyFunctionSpecification()
        function.addParameter('comm_identifier', dtype='int32', direction=function.IN)
        function.addParameter('encoded_interface', dtype='string', direction=function.IN)
        function.result_type = 'int32'
        function.can_handle_array = False
        return function

    @legacy_function
    def deep_echo_string():
        function = LegacyFunctionSpecification()
        function.addParameter('string_in', dtype='string', direction=function.IN)
        function.addParameter('string_out', dtype='string', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def return_control():
        function = LegacyFunctionSpecification()
        function.result_type = 'int32'
        function.can_handle_array = False
        return function

    @legacy_function
    def echo_quantity():
        function = LegacyFunctionSpecification()
        function.addParameter('quantity_in', dtype='float64', direction=function.IN)
        function.addParameter('quantity_out', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        function.has_units = True
        function.id = 23
        return function

    @legacy_function
    def echo_quantities():
        function = LegacyFunctionSpecification()
        function.addParameter('quantity_in', dtype='float64', direction=function.IN)
        function.addParameter('quantity_out', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.must_handle_array = True
        function.has_units = True
        function.id = 23
        return function

    @legacy_function
    def echo_quantities_error():
        function = LegacyFunctionSpecification()
        function.addParameter('quantity_in', dtype='float64', direction=function.IN)
        function.addParameter('quantity_out', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.must_handle_array = True
        function.has_units = True
        function.id = 25
        return function


basic_python_exe = """#!{executable}
import sys
import os
from subprocess import call

if __name__ == '__main__':
    command = sys.argv[1:]
    
    dirname=os.path.dirname(__file__)
    dirname=os.path.abspath(dirname)
    
    logfile=os.path.join(dirname, 'pythonexe.log')
    
    with open(logfile, 'a') as stream:
        stream.write('start {{0}}\\n'.format(command[0]))
        stream.flush()
    
    command.insert(0, sys.executable)
    
    returncode = call(command, close_fds=False)
        
    with open(logfile, 'a') as stream:
        stream.write('end {{0}} {{1}}\\n'.format(command[0], returncode))
        stream.flush()
        
    sys.exit(returncode)
"""


class ForTestingImplementation(object):

    def __init__(self):
        self.masses = [0.0] * 100
        self._particle_data = numpy.reshape(numpy.arange(300.0), (-1, 3))
        self.maxindex = 0

    def new_particle(self,  mass, other, index_of_the_particle):
        try:
            self.masses[self.maxindex] = mass
            index_of_the_particle.value = self.maxindex
            self.maxindex += 1
            return 0
        except:
            return -1

    def get_mass(self, index_of_the_particle,  mass):
        try:
            mass.value = self.masses[index_of_the_particle]
            return 0
        except:
            return -1

    def set_mass(self, index_of_the_particle,  mass):
        try:
            self.masses[index_of_the_particle] = mass
            return 0
        except:
            return -1

    def echo_int(self, int_in, int_out):
        int_out.value = int_in
        return 0

    def echo_double(self, double_in, double_out):
        double_out.value = double_in
        return 0

    def echo_string(self, string_in, string_out):
        string_out.value = string_in
        return 0

    def echo_bool(self, in_, out_):
        out_.value = in_
        return 0

    def print_string(self, string_in):
        print(string_in)
        return 0

    def print_error_string(self, string_in):
        print(string_in, file=sys.stderr)
        return 0

    def echo_strings(self, string_inout1, string_inout2):
        string_inout1.value = string_inout1.value[::-1]
        string_inout2.value = string_inout2.value[::-1]
        return 0

    def sleep(self, number_of_seconds):
        import time
        time.sleep(number_of_seconds)
        return 0

    def sum_doubles(self, double_in1, double_in2, double_out):
        double_out.value = double_in1 + double_in2
        return 0

    def multiply_ints(self, int_in1, int_in2, int_out):
        int_out.value = int_in1 * int_in2
        return 0

    def get_position(self, index_of_the_particle, x, y, z, length):
        try:
            x.value = self._particle_data[index_of_the_particle, 0]
            y.value = self._particle_data[index_of_the_particle, 1]
            z.value = self._particle_data[index_of_the_particle, 2]
            return 0
        except:
            return -1

    def set_position(self, index_of_the_particle, x, y, z, length):
        try:
            self._particle_data[index_of_the_particle, 0] = x
            self._particle_data[index_of_the_particle, 1] = y
            self._particle_data[index_of_the_particle, 2] = z
            return 0
        except:
            return -1

    def copy_over_interface(self, comm_identifier, encoded_interface):
        self._other = pickle.loads(encoded_interface.encode("latin-1"))
        self._other.channel.intercomm = self._interface.communicators[comm_identifier]
        return 0

    def deep_echo_string(self, string_in, string_out):
        result, errorcode = self._other.echo_string(string_in)
        string_out.value = result[::-1]
        return errorcode

    def return_control(self):
        self._other.internal__activate_communicator(0)
        return 0

    def echo_quantity(self, quantity_in, quantity_out):
        quantity_out.value = quantity_in * (10 | (1.0/units.s))
        return 0

    def echo_quantities(self, quantity_in, quantity_out):
        quantity_out.value = quantity_in * (10 | (1.0/units.s))
        return 0

    def echo_quantities_error(self, quantity_in, quantity_out):
        raise Exception("an unexpected event")
        return 0


class ForTesting(InCodeComponentImplementation):

    def __init__(self, **options):
        InCodeComponentImplementation.__init__(self, ForTestingInterface(**options), **options)

    def define_methods(self, object):
        object.add_method("sleep", (units.s,), (object.ERROR_CODE,))
        object.add_method("new_particle", (units.kg, object.LINK("particles")), (object.INDEX, object.ERROR_CODE,))

    def define_particle_sets(self, object):
        object.define_set('particles', 'index_of_the_particle')
        object.set_new('particles', 'new_particle')

        object.set_delete('particles', 'delete_particle')
        object.add_setter('particles', 'set_mass')
        object.add_getter('particles', 'get_mass')


class TestCreatePythonWorker(TestCase):

    def test1(self):
        x = CreateAPythonWorker()
        x.implementation_factory = ForTestingImplementation
        x.channel_type = 'mpi'
        x.interface_class = ForTestingInterface

        script_string = x.new_executable_script_string()
        self.assertTrue(script_string.find('syspath = (') > 0)
        self.assertTrue(script_string.find('ForTestingInterface') > 0)
        self.assertTrue(script_string.find('ForTestingImplementation') > 0)
        self.assertTrue(script_string.find('test_python_implementation') > 0)
        self.assertTrue(script_string.find('PythonImplementation(instance, ForTestingInterface)') > 0)
        try:
            st = compile(script_string, 'test.py', 'exec')
        except SyntaxError as ex:
            self.fail("Compilation error {0}".format(ex))

    def test2(self):
        x = CreateAPythonWorker()
        x.implementation_factory = ForTestingImplementation
        x.channel_type = 'sockets'
        x.interface_class = ForTestingInterface

        script_string = x.new_executable_script_string()
        self.assertTrue(script_string.find('syspath = (') > 0)
        self.assertTrue(script_string.find('ForTestingInterface') > 0)
        self.assertTrue(script_string.find('ForTestingImplementation') > 0)
        self.assertTrue(script_string.find('test_python_implementation') > 0)
        self.assertTrue(script_string.find('PythonImplementation(instance, ForTestingInterface)') > 0)
        self.assertTrue(script_string.find('start_socket') > 0)
        try:
            st = compile(script_string, 'test.py', 'exec')
        except SyntaxError as ex:
            self.fail("Compilation error {0}".format(ex))


class TestInterface(TestWithMPI):

    def ForTesting(self, **options):
        options["worker_dir"] = self.get_path_to_results()
        return ForTesting(**options)

    def ForTestingInterface(self, **options):
        options["worker_dir"] = self.get_path_to_results()
        return ForTestingInterface(**options)

    def test02(self):
        implementation = ForTestingImplementation()
        x = python_code.PythonImplementation(implementation, ForTestingInterface)

        input_message = python_code.ClientSideMPIMessage(0, 10, 1)
        input_message.ints = [1]

        output_message = python_code.ClientSideMPIMessage(0, 10, 1)

        x.handle_message(input_message, output_message)

        self.assertEqual(len(output_message.ints), 1)
        self.assertEqual(len(output_message.doubles), 1)
        self.assertEqual(output_message.ints[0], 0)
        self.assertEqual(output_message.doubles[0], 0.0)

    def test03(self):
        implementation = ForTestingImplementation()
        x = python_code.PythonImplementation(implementation, ForTestingInterface)

        input_message = python_code.ClientSideMPIMessage(0, 11, 1)
        input_message.ints = [1]
        input_message.doubles = [12.0]

        output_message = python_code.ClientSideMPIMessage(0, 10, 1)

        x.handle_message(input_message, output_message)

        self.assertEqual(len(output_message.ints), 1)
        self.assertEqual(len(output_message.doubles), 0)
        self.assertEqual(output_message.ints[0], 0)
        self.assertEqual(implementation.masses[1], 12.0)

    def test04(self):
        implementation = ForTestingImplementation()
        x = python_code.PythonImplementation(implementation, ForTestingInterface)

        input_message = python_code.ClientSideMPIMessage(0, 11, 4)
        input_message.ints = [1, 2, 3, 4]
        input_message.doubles = [12.0, 13.0, 14.0, 15.0]

        output_message = python_code.ClientSideMPIMessage(0, 10, 4)

        x.handle_message(input_message, output_message)

        self.assertEqual(len(output_message.ints), 4)
        self.assertEqual(len(output_message.doubles), 0)
        self.assertEqual(output_message.ints[0], 0)
        self.assertEqual(output_message.ints[3], 0)
        self.assertEqual(implementation.masses[1], 12.0)
        self.assertEqual(implementation.masses[2], 13.0)
        self.assertEqual(implementation.masses[3], 14.0)
        self.assertEqual(implementation.masses[4], 15.0)

    def test05(self):
        x = self.ForTestingInterface()

        error = x.set_mass(1, 10.0)
        self.assertEqual(error, 0)

        answer, error = x.get_mass(1)
        self.assertEqual(error, 0)
        self.assertEqual(answer, 10.0)
        x.stop()

    def test06(self):
        x = self.ForTestingInterface()

        errors = x.set_mass([1, 2], [10.0, 11.0])
        self.assertEqual(errors[0], 0)
        self.assertEqual(errors[1], 0)

        answer, errors = x.get_mass([1, 2])
        self.assertEqual(errors[0], 0)
        self.assertEqual(answer[0], 10.0)
        self.assertEqual(answer[1], 11.0)
        x.stop()

        x.stop()

    def test07(self):
        x = self.ForTestingInterface()

        int_out, error = x.echo_int(20)
        self.assertEqual(error, 0)
        self.assertEqual(int_out, 20)
        x.stop()

        x.stop()

    def test08(self):
        implementation = ForTestingImplementation()
        x = python_code.PythonImplementation(implementation, ForTestingInterface)

        input_message = python_code.ClientSideMPIMessage(0, 12, 1)
        input_message.ints = [20]

        output_message = python_code.ClientSideMPIMessage(0, 10, 1)

        x.handle_message(input_message, output_message)

        self.assertEqual(len(output_message.ints), 2)
        self.assertEqual(output_message.ints[0], 0)
        self.assertEqual(output_message.ints[1], 20)

    def test09(self):
        x = self.ForTestingInterface()
        string_out, error = x.echo_string("1234567")
        self.assertEqual(error, 0)
        self.assertEqual(string_out, "1234567")
        x.stop()

    def test10(self):
        x = self.ForTestingInterface()
        string_out, error = x.echo_string(["aaaaa", "bbbb"])
        self.assertEqual(error[0], 0)
        self.assertEqual(len(string_out), 2)
        self.assertEqual(string_out[0], "aaaaa")
        self.assertEqual(string_out[1], "bbbb")
        x.stop()

    def test11(self):
        x = self.ForTestingInterface()
        string_out, error = x.echo_string(["", "bbbb"])
        self.assertEqual(error[0], 0)
        self.assertEqual(len(string_out), 2)
        self.assertEqual(string_out[0], "")
        self.assertEqual(string_out[1], "bbbb")
        x.stop()

    def test12(self):
        x = self.ForTestingInterface()
        str1_out, str2_out, error = x.echo_strings("abc", "def")
        self.assertEqual(error, 0)
        self.assertEqual(str1_out, "cba")
        self.assertEqual(str2_out, "fed")
        x.stop()

    def test13(self):
        x = self.ForTestingInterface()
        str1_out, str2_out, error = x.echo_strings(["abc", "def"], ["ghi", "jkl"])
        self.assertEqual(error[0], 0)
        self.assertEqual(error[1], 0)
        self.assertEqual(str1_out[0], "cba")
        self.assertEqual(str1_out[1], "fed")
        self.assertEqual(str2_out[0], "ihg")
        self.assertEqual(str2_out[1], "lkj")
        x.stop()

    def test14(self):
        x = self.ForTestingInterface()
        result = x.sleep(2)
        self.assertEqual(result, 0)
        request = x.sleep.asynchronous(0.01)
        request.wait()
        result = request.result()
        self.assertEqual(result, 0)
        x.stop()

    def test15(self):
        x = self.ForTestingInterface()
        y = self.ForTestingInterface()
        request1 = x.sleep.asynchronous(0.5)
        self.assertFalse(request1.is_result_available())
        request2 = y.sleep.asynchronous(1.5)
        self.assertFalse(request2.is_result_available())
        request2.wait()
        self.assertTrue(request1.is_result_available())
        self.assertTrue(request2.is_result_available())

        self.assertEqual(request1.result(), 0)
        self.assertEqual(request2.result(), 0)

        y.stop()
        x.stop()

    def test16(self):
        x = self.ForTestingInterface()
        request1 = x.sleep.asynchronous(0.4)
        x.sleep(0.01)
        self.assertTrue(request1.is_result_available(), True)
        request1.wait()
        self.assertTrue(request1.is_result_available())
        x.sleep(0.01)
        self.assertTrue(request1.is_result_available())
        x.stop()

    def test17(self):
        x = self.ForTesting()
        self.assertTrue(x.sleep.is_async_supported)
        request = x.sleep.asynchronous(0.2 | units.s)
        request.wait()
        result = request.result()
        self.assertEqual(result, [])
        x.stop()

    def test18(self):
        print("Testing the splitting of very long MPI messages into blocks")
        x = self.ForTesting(max_message_length=10)
        N = 100
        doubles = x.echo_double([1.0*i for i in range(N)])
        self.assertTrue(list(doubles) == [1.0*i for i in range(N)])
        sums = x.sum_doubles([1.0*i for i in range(N)], [1.0*i for i in range(N)])
        self.assertTrue(list(sums) == [2.0*i for i in range(N)])
        products = x.multiply_ints(range(N), range(N))
        self.assertTrue(list(products) == [i*i for i in range(N)])
        N = 101
        doubles = x.echo_double([1.0*i for i in range(N)])
        self.assertTrue(list(doubles) == [1.0*i for i in range(N)])
        sums = x.sum_doubles([1.0*i for i in range(N)], [1.0*i for i in range(N)])
        self.assertTrue(list(sums) == [2.0*i for i in range(N)])
        products = x.multiply_ints(range(N), range(N))
        self.assertTrue(list(products) == [i*i for i in range(N)])
        x.stop()

    def test19(self):
        print("Testing the splitting of very long MPI messages into blocks II: strings")
        x = self.ForTesting(max_message_length=10)
        N = 100
        strings1, strings2 = x.echo_strings(['REDRUM' for i in range(N)], ['stressed' for i in range(N)])
        self.assertTrue(list(strings1) == ['MURDER' for i in range(N)])
        self.assertTrue(list(strings2) == ['desserts' for i in range(N)])
        N = 101
        strings1, strings2 = x.echo_strings(['REDRUM' for i in range(N)], ['stressed' for i in range(N)])
        self.assertTrue(list(strings1) == ['MURDER' for i in range(N)])
        self.assertTrue(list(strings2) == ['desserts' for i in range(N)])
        x.stop()

    def xtest20(self):
        #
        # TURNED OFF support for redirection,
        # by default output is redirected to /dev/null
        # if you need file, use the support from your mpi implementation
        #
        if os.path.exists("pout.000"):
            os.remove("pout.000")
        if os.path.exists("perr.000"):
            os.remove("perr.000")

        x = self.ForTesting(redirect_stderr_file='perr', redirect_stdout_file='pout', redirection="file")
        x.print_string("abc")
        x.print_error_string("exex")
        x.stop()

        time.sleep(0.2)

        self.assertTrue(os.path.exists("pout.000"))
        with open("pout.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "abc")

        self.assertTrue(os.path.exists("perr.000"))
        with open("perr.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "exex")

        x = self.ForTesting(redirect_stderr_file='pout', redirect_stdout_file='pout', redirection="file")
        x.print_string("def")
        x.print_error_string("exex")
        x.stop()

        time.sleep(0.2)

        self.assertTrue(os.path.exists("pout.000"))
        with open("pout.000", "r") as f:
            content = f.read()
        self.assertEqual(content.strip(), "abc\ndef\nexex")

    def test21(self):
        print("Testing must_handle_array for Python codes")
        instance = self.ForTestingInterface()

        x, y, z, err = instance.get_position(range(100))
        self.assertEqual(err, 0)
        self.assertEqual(x, numpy.arange(0.0, 300.0, 3.0))
        self.assertEqual(y, numpy.arange(1.0, 300.0, 3.0))
        self.assertEqual(z, numpy.arange(2.0, 300.0, 3.0))
        x, y, z, err = instance.get_position(range(101))
        self.assertEqual(err, -1)
        self.assertEqual(list(instance.get_position(1).values()), [3.0, 4.0, 5.0, 0])

        err = instance.set_position(range(100), numpy.arange(100.0), numpy.arange(100.0, 200.0), numpy.arange(200.0, 300.0))
        self.assertEqual(err, 0)
        err = instance.set_position(range(101), numpy.arange(101.0), numpy.arange(101.0, 202.0), numpy.arange(202.0, 303.0))
        self.assertEqual(err, -1)
        err = instance.set_position(0, -1.0, -2.0, -3.0)

        x, y, z, err = instance.get_position(range(100))
        self.assertEqual(err, 0)
        self.assertEqual(x, numpy.concatenate(([-1.0], numpy.arange(1.0, 100.0))))
        self.assertEqual(y, numpy.concatenate(([-2.0], numpy.arange(101.0, 200.0))))
        self.assertEqual(z, numpy.concatenate(([-3.0], numpy.arange(201.0, 300.0))))

        instance.stop()

    def test22(self):

        pool = AsyncRequestsPool()

        x = self.ForTestingInterface()
        y = self.ForTestingInterface()
        request1 = x.sleep.asynchronous(0.5)
        request2 = y.sleep.asynchronous(1.5)
        finished_requests = []

        def handle_result(request, index):
            self.assertTrue(request.is_result_available())
            finished_requests.append(index)

        pool.add_request(request1, handle_result, [1])
        pool.add_request(request2, handle_result, [2])

        pool.wait()
        self.assertEqual(len(finished_requests), 1)
        self.assertEqual(len(pool), 1)

        pool.wait()
        self.assertEqual(len(finished_requests), 2)
        self.assertEqual(len(pool), 0)

        self.assertTrue(request1.is_result_available())
        self.assertTrue(request2.is_result_available())

        self.assertEqual(request1.result(), 0)
        self.assertEqual(request2.result(), 0)

        y.stop()
        x.stop()

    def test22b(self):

        pool = AsyncRequestsPool()

        x = self.ForTestingInterface()
        y = self.ForTestingInterface()
        request1 = x.sleep.asynchronous(0.2)
        request2 = y.sleep.asynchronous(0.2)
        finished_requests = []

        def handle_result(request, index):
            self.assertTrue(request.is_result_available())
            finished_requests.append(index)

        pool.add_request(request1, handle_result, [1])
        pool.add_request(request2, handle_result, [2])

        time.sleep(1.0)

        pool.wait()
        pool.wait()

        self.assertEqual(len(finished_requests), 2)
        self.assertEqual(len(pool), 0)

        self.assertTrue(request1.is_result_available())
        self.assertTrue(request2.is_result_available())

        self.assertEqual(request1.result(), 0)
        self.assertEqual(request2.result(), 0)

        pool.wait()
        self.assertEqual(len(pool), 0)

        y.stop()
        x.stop()

    def test23(self):

        path = self.get_path_to_results()

        exe = os.path.join(path, "pythonexe")
        log = os.path.join(path, "pythonexe.log")

        if os.path.exists(exe):
            os.remove(exe)

        if os.path.exists(log):
            os.remove(log)

        string = basic_python_exe.format(executable=sys.executable)

        with open(exe, 'w') as f:
            f.write(string)

        os.chmod(exe, 0o777)

        instance = self.ForTestingInterface(
            use_python_interpreter=True,
            python_interpreter=exe
        )
        x, y, z, err = instance.get_position(range(100))
        self.assertEqual(err, 0)
        self.assertEqual(x, numpy.arange(0.0, 300.0, 3.0))
        self.assertEqual(y, numpy.arange(1.0, 300.0, 3.0))
        self.assertEqual(z, numpy.arange(2.0, 300.0, 3.0))

        instance.stop()
        # this is quite long, but apparently necessary
        time.sleep(2.)

        self.assertTrue(os.path.exists(log))

        with open(log, 'r') as f:
            loglines = f.read().splitlines()
        print(loglines)

        self.assertEqual(len(loglines), 2)

        self.assertTrue(loglines[0].startswith('start '))
        self.assertTrue(loglines[1].startswith('end '))

    def test24(self):

        # same as test23 but now with redirection is none
        path = self.get_path_to_results()

        exe = os.path.join(path, "pythonexe")
        log = os.path.join(path, "pythonexe.log")

        if os.path.exists(exe):
            os.remove(exe)

        if os.path.exists(log):
            os.remove(log)

        string = basic_python_exe.format(executable=sys.executable)

        with open(exe, 'w') as f:
            f.write(string)

        os.chmod(exe, 0o777)

        instance = self.ForTestingInterface(
            use_python_interpreter=True,
            python_interpreter=exe,
            redirection="none"
        )
        x, y, z, err = instance.get_position(range(100))
        self.assertEqual(err, 0)
        self.assertEqual(x, numpy.arange(0.0, 300.0, 3.0))
        self.assertEqual(y, numpy.arange(1.0, 300.0, 3.0))
        self.assertEqual(z, numpy.arange(2.0, 300.0, 3.0))

        instance.stop()
        time.sleep(2.)

        self.assertTrue(os.path.exists(log))

        with open(log, 'r') as f:
            loglines = f.read().splitlines()

        self.assertEqual(len(loglines), 2)

        self.assertTrue(loglines[0].startswith('start '))
        self.assertTrue(loglines[1].startswith('end '))

    def test25(self):
        self.check_for_mpi()
        instance = self.ForTestingInterface(polling_interval_in_milliseconds=100)
        (output1, error1) = instance.internal__get_message_polling_interval()
        instance.stop()
        self.assertEqual(error1, 0)
        self.assertEqual(output1, 100000)

    def test25(self):
        instance = self.ForTestingInterface(polling_interval_in_milliseconds=100)
        if instance.channel.is_polling_supported():
            (output1, error1) = instance.internal__get_message_polling_interval()
            self.assertEqual(error1, 0)
            self.assertEqual(output1, 100000)
        instance.stop()

# these test might need special configuration, e.g. for openmpi:
# run "ompi-server --no-daemonize -r -"
# and set OMPI_MCA_pmix_server_uri="<uri>" with <uri> the reported uri
    def test26(self):
        self.check_for_mpi()
        instance1 = self.ForTestingInterface()
        instance2 = self.ForTestingInterface()
        portname, error = instance1.internal__open_port()
        self.assertTrue(len(portname) > 0)
        self.assertEqual(error, 0)
        request1 = instance1.internal__accept_on_port.asynchronous(portname)
        request2 = instance2.internal__connect_to_port.asynchronous(portname)
        request1.wait()
        request2.wait()
        port_id1, error1 = request1.result()
        port_id2, error2 = request2.result()
        self.assertTrue(port_id1 >= 0)
        self.assertTrue(port_id2 >= 0)
        self.assertEqual(error1, 0)
        self.assertEqual(error2, 0)

# this test requires that instance1 and instance2 use MPIchannel
# note shutdown of remote process triggers (ignored) invalid communicator error
    def test27(self):
        self.check_for_mpi()
        instance1 = self.ForTestingInterface(redirection="none")
        instance2 = self.ForTestingInterface(redirection="none")
        encoded_interface = pickle.dumps(instance1, 0)
        decoded_interface = pickle.loads(encoded_interface)
        # pickle.loads(pickle.dumps(instance1,0))
        portname, error = instance2.internal__open_port()
        request1 = instance2.internal__accept_on_port.asynchronous(portname)
        request2 = instance1.internal__connect_to_port.asynchronous(portname)
        request1.wait()
        request2.wait()
        port_id1, error1 = request1.result()
        port_id2, error2 = request2.result()
        instance2.copy_over_interface(port_id2, pickle.dumps(instance1, 0).decode('latin-1'))
        instance1.internal__activate_communicator(port_id1)
        result, errorcode = instance2.deep_echo_string("hello")
        self.assertEqual(errorcode, 0)
        self.assertEqual(result, "olleh")
        result, errorcode = instance2.deep_echo_string("world")
        self.assertEqual(errorcode, 0)
        self.assertEqual(result, "dlrow")
        instance2.return_control()
        result, errorcode = instance1.echo_string("world")
        self.assertEqual(errorcode, 0)
        self.assertEqual(result, "world")
        instance1.stop()
        instance2.stop()

    def test28(self):
        x = self.ForTestingInterface()

        def next_request(index):
            if index < 3:
                return x.sleep.asynchronous(0.1)
            else:
                return None

        sequence = ASyncRequestSequence(next_request)
        self.assertFalse(sequence.is_finished)
        sequence.wait()
        self.assertTrue(sequence.is_finished)
        result = sequence.result()
        self.assertEqual(len(result), 3)
        x.stop()

    def test29(self):

        pool = AsyncRequestsPool()

        x = self.ForTestingInterface()
        y = self.ForTestingInterface()
        sequenced_requests_indices = []

        def next_request(index):
            if index < 4:
                sequenced_requests_indices.append(index)
                return x.sleep.asynchronous(0.5)
            else:
                return None

        request1 = ASyncRequestSequence(next_request)
        request2 = y.sleep.asynchronous(1.0)
        finished_requests = []

        def handle_result(request, index):
            self.assertTrue(request.is_result_available())
            self.assertTrue(request.is_finished)
            finished_requests.append(index)

        pool.add_request(request1, handle_result, [1])
        pool.add_request(request2, handle_result, [2])

        pool.wait()
        self.assertEqual(len(finished_requests), 1)
        self.assertEqual(len(pool), 1)
        self.assertEqual(finished_requests, [2])
        self.assertTrue(len(sequenced_requests_indices) > 0)

        pool.wait()
        self.assertEqual(len(finished_requests), 2)
        self.assertEqual(len(pool), 0)
        x.sleep(0.1)
        self.assertEqual(sequenced_requests_indices, [0, 1, 2, 3])

        self.assertTrue(request1.is_result_available())
        self.assertTrue(request2.is_result_available())

        self.assertEqual(request1.result(), [0, 0, 0, 0])
        self.assertEqual(request2.result(), 0)

        y.stop()
        x.stop()

    def test30(self):
        instance = self.ForTesting()
        input = [1.0, 2.0, 3.0]
        output = instance.sum_doubles(input, 5)
        self.assertAlmostRelativeEquals(output, [6.0, 7.0, 8.0])
        output = instance.sum_doubles(5, input)
        self.assertAlmostRelativeEquals(output, [6.0, 7.0, 8.0])

    def test31(self):
        x = self.ForTesting()
        p = datamodel.Particles(5)
        p.mass = [1, 2, 3, 4, 5] | units.kg
        p.other = None
        for pi in p:
            x.particles.add_particle(pi)
        self.assertAlmostRelativeEquals(x.particles.mass, [1, 2, 3, 4, 5])
        x.stop()

    def test32(self):
        x = self.ForTestingInterface()
        quantity_out, error = x.echo_quantity(20.0 | units.m)
        self.assertEqual(error, 0)
        self.assertEqual(quantity_out, 200 | (units.m/units.s))
        quantity_out, error = x.echo_quantity(30)
        self.assertEqual(error, 0)
        self.assertEqual(quantity_out, 300 | (1.0/units.s))
        x.stop()

    def test33(self):
        x = self.ForTestingInterface()
        quantity_out, error = x.echo_quantity([20, 30, 40] | units.m)
        self.assertEqual(error, 0)
        self.assertEqual(quantity_out, [200, 300, 400] | (units.m/units.s))
        x.stop()

    def test34(self):
        x = self.ForTestingInterface()
        # self.assertException(x.echo_quantities_error, [20, 30, 40] | units.m)
        quantity_out, error = x.echo_quantities([20, 30, 40] | units.m)
        self.assertEqual(error, 0)
        self.assertEqual(quantity_out, [200, 300, 400] | (units.m/units.s))
        x.stop()

    def test35(self):
        x = self.ForTesting(max_message_length=10)
        N = 10
        doubles = x.echo_double([1.0*i for i in range(N)])
        self.assertTrue(list(doubles) == [1.0*i for i in range(N)])
        sums = x.sum_doubles([3.0*i for i in range(N)])
        self.assertTrue(list(sums) == [3.0*i + 1 for i in range(N)])
        N = 11
        doubles = x.echo_double([1.0*i for i in range(N)])
        self.assertTrue(list(doubles) == [1.0*i for i in range(N)])
        sums = x.sum_doubles([3.0*i for i in range(N)])
        self.assertTrue(list(sums) == [3.0*i + 1 for i in range(N)])
        x.stop()

    def test36(self):
        x = self.ForTestingInterface()
        self.assertRaises(exceptions.CodeException, x.echo_quantities_error, ([20, 30, 40] | units.m), expected_message="Exception when calling function 'echo_quantities_error', of code 'ForTestingInterface', exception was 'Error in code: an unexpected event'")
        x.stop()

    def test37(self):
        x = self.ForTestingInterface()
        request = x.echo_quantity.asynchronous([20, 30, 40] | units.m)
        quantity_out, error = request.result()
        self.assertEqual(error, 0)
        self.assertEqual(quantity_out, [200, 300, 400] | (units.m/units.s))
        x.stop()

    def test40(self):
        x = self.ForTesting()
        out = x.echo_bool([True, False, True])
        self.assertEqual(out, [True, False, True])
        x.stop()



import numpy

from amuse.test import amusetest
from amuse.support.exceptions import AmuseException
from amuse.units import nbody_system
from amuse.units import units
from amuse.ic.kingmodel import new_king_model


class TestKingModel(amusetest.TestCase):
    def test1(self):
        print("First test: making a King model.")
        number_of_particles = 10
        particles = new_king_model(number_of_particles, 6.0)
        self.assertAlmostEqual(particles.mass.sum(), 1 | nbody_system.mass)
        self.assertAlmostEqual(particles[0].mass, 0.1 | nbody_system.mass)
        print(particles)
        self.assertFalse(particles[0].mass.is_vector())

    def test2(self):
        print("Testing kinetic and potential energy of a King model realisation.")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        number_of_particles = 500
        particles = new_king_model(number_of_particles, 6.0, convert_nbody, do_scale=True)
        self.assertEqual(len(particles), number_of_particles)
        self.assertAlmostEqual(particles[0].mass, (1.0 | units.MSun)/number_of_particles, 3, in_units=units.MSun)
        self.assertAlmostEqual(convert_nbody.to_nbody(particles.kinetic_energy()), 0.25 | nbody_system.energy)
        self.assertAlmostEqual(convert_nbody.to_nbody(particles.potential_energy()), -0.5 | nbody_system.energy)

    def slowtest3(self):
        print("King models with varying King dimensionless depth W0.")
        number_of_particles = 10
        for w_0 in [1.0, 6.0, 11.0, 16.0]:
            particles = new_king_model(number_of_particles, W0=w_0)
            self.assertEqual(len(particles), number_of_particles)

    def test4(self):
        print("Testing maximal/minimal value of King dimensionless depth W0.")
        number_of_particles = 10
        self.assertRaises(AmuseException, new_king_model, number_of_particles, W0=16.5,
            expected_message="makeking: must specify w0 < 16")

        self.assertRaises(ZeroDivisionError, new_king_model, number_of_particles, W0=0.0)

    def test5(self):
        print("Testing a specific King model realisation.")
        numpy.random.seed(345672)
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        particles = new_king_model(500, 6.0, convert_nbody)
        self.assertEqual(len(particles), 500)
        self.assertAlmostEqual(particles.total_mass(), 1.0 | units.MSun)
        self.assertAlmostEqual(particles.mass, 1.0 / 500 | units.MSun)
        self.assertAlmostEqual(particles.center_of_mass(),          [0, 0, 0] | units.AU)
        self.assertAlmostEqual(particles.center_of_mass_velocity(), [0, 0, 0] | units.km / units.s)
        self.assertAlmostEqual(particles[:3].position, [[-0.23147381, -0.19421449, -0.01165137],
            [-0.09283025, -0.06444658, -0.07922396], [-0.44189946, 0.23786357, 0.39115629]] | units.AU)



import os
import os.path
import numpy
import platform
from amuse.community import *
from amuse.test.amusetest import TestWithMPI

from amuse.community.galactics.interface import GalactICsInterface, GalactICs

# Change the default for some GalactICs(-Interface) keyword arguments:
default_options = dict()
# default_options = dict(redirection = "none")


class GalactICsInterfaceTests(TestWithMPI):

    def test1(self):
        print("Testing GalactICsInterface initialization")
        instance = GalactICsInterface(**default_options)
        self.assertEqual(instance.initialize_code(), 0)
        self.assertEqual(instance.set_output_path(instance.get_output_directory()), 0)
        self.assertEqual(instance.set_generate_bulge_flag(False), 0)
        self.assertEqual(instance.set_generate_disk_flag(False), 0)
        self.assertEqual(instance.set_order_of_multipole_expansion(0), 0)
        self.assertEqual(instance.commit_parameters(), 0)
        self.assertEqual(instance.cleanup_code(), 0)
        instance.stop()

    def test2(self):
        print("Testing GalactICsInterface parameters")
        instance = GalactICsInterface(**default_options)
        self.assertEqual(instance.initialize_code(), 0)
        self.assertEqual(instance.set_output_path(os.path.join(instance.get_output_directory(), "test")), 0)

        self.assertEqual(instance.set_generate_halo_flag(False), 0)
        self.assertEqual(instance.set_disk_do_center_flag(False), 0)
        self.assertEqual(instance.set_number_of_grid_intervals(50000), 0)
        self.assertEqual(instance.set_disk_random_seed(-1234), 0)
        self.assertEqual(instance.set_halo_outer_radius(250.0), 0)
        self.assertEqual(instance.set_bulge_streaming_fraction(0.4), 0)

        self.assertEqual([False, 0], list(instance.get_generate_halo_flag().values()))
        self.assertEqual([False, 0], list(instance.get_disk_do_center_flag().values()))
        self.assertEqual([50000, 0], list(instance.get_number_of_grid_intervals().values()))
        self.assertEqual([-1234, 0], list(instance.get_disk_random_seed().values()))
        self.assertEqual([250.0, 0], list(instance.get_halo_outer_radius().values()))
        self.assertEqual([0.4, 0], list(instance.get_bulge_streaming_fraction().values()))

        self.assertEqual(instance.cleanup_code(), 0)
        instance.stop()

    def slowtest3(self):
        print("Testing GalactICsInterface generate_particles")
        n_particles_halo = 100
        n_particles_bulge = 100
        n_particles_disk = 100
        number_of_particles = n_particles_disk + n_particles_bulge + n_particles_halo

        instance = GalactICsInterface(**default_options)
        self.assertEqual(instance.initialize_code(), 0)
        self.assertEqual(instance.set_output_path(instance.get_output_directory()), 0)
        self.assertEqual(instance.set_halo_number_of_particles(n_particles_halo), 0)
        self.assertEqual(instance.set_bulge_number_of_particles(n_particles_bulge), 0)
        self.assertEqual(instance.set_disk_number_of_particles(n_particles_disk), 0)
        self.assertEqual(instance.commit_parameters(), 0)

        self.assertEqual(list(instance.get_number_of_particles_updated().values()), [0, 0])
        self.assertEqual(instance.generate_particles(), 0)
        self.assertEqual(list(instance.get_number_of_particles_updated().values()), [number_of_particles, 0])

        mass_disk, mass_bulge, mass_halo = 26.578816771507263, 14.632800221443176, 1184.2350006103516
        masses, errors = instance.get_mass(range(number_of_particles))
        self.assertEqual(errors, numpy.zeros(number_of_particles))
        self.assertAlmostRelativeEquals(masses, numpy.concatenate((
            numpy.ones(n_particles_disk)*mass_disk/n_particles_disk,
            numpy.ones(n_particles_bulge)*mass_bulge/n_particles_bulge,
            numpy.ones(n_particles_halo)*mass_halo/n_particles_halo,
        )), 3)

        x_positions, y_positions, z_positions, errors = instance.get_position(range(number_of_particles))
        self.assertEqual(errors, numpy.zeros(number_of_particles))
        self.assertAlmostEqual(numpy.array([numpy.mean(x_positions), numpy.mean(y_positions),
            numpy.mean(z_positions)]), numpy.array([0.0]*3), 5)
        self.assertAlmostRelativeEquals(numpy.array([
            numpy.mean(abs(x_positions[:n_particles_disk])),
            numpy.mean(abs(y_positions[:n_particles_disk])),
            numpy.mean(abs(z_positions[:n_particles_disk]))]),
            numpy.array([7.3994484072923656, 7.1570298135280606, 0.33854196755215527]), 3)
        self.assertAlmostRelativeEquals(numpy.array([
            numpy.mean(abs(x_positions[n_particles_disk:n_particles_disk+n_particles_bulge])),
            numpy.mean(abs(y_positions[n_particles_disk:n_particles_disk+n_particles_bulge])),
            numpy.mean(abs(z_positions[n_particles_disk:n_particles_disk+n_particles_bulge]))]),
            numpy.array([1.244429082274437, 1.1639373835548759, 0.8550614269822836]), 3)
        self.assertAlmostRelativeEquals(numpy.array([
            numpy.mean(abs(x_positions[-n_particles_halo:])),
            numpy.mean(abs(y_positions[-n_particles_halo:])),
            numpy.mean(abs(z_positions[-n_particles_halo:]))]),
            numpy.array([94.242819476127622, 88.41320479869843, 85.234394512176507]), 3)

        x_velocities, y_velocities, z_velocities, errors = instance.get_velocity(range(number_of_particles))
        self.assertEqual(errors, numpy.zeros(number_of_particles))
        self.assertAlmostEqual(numpy.array([numpy.mean(x_velocities), numpy.mean(y_velocities),
            numpy.mean(z_velocities)]), numpy.array([0.0]*3))
        self.assertAlmostRelativeEquals(numpy.array([
            numpy.mean(abs(x_velocities[:n_particles_disk])),
            numpy.mean(abs(y_velocities[:n_particles_disk])),
            numpy.mean(abs(z_velocities[:n_particles_disk]))]),
            numpy.array([1.5026254250109197, 1.5649469271302223, 0.20230436498299242]), 5)
        self.assertAlmostRelativeEquals(numpy.array([
            numpy.mean(abs(x_velocities[n_particles_disk:])),
            numpy.mean(abs(y_velocities[n_particles_disk:])),
            numpy.mean(abs(z_velocities[n_particles_disk:]))]),
            numpy.array([0.99470628838986164, 0.95913934175856408, 0.9359876788407564]), 5)

        self.assertEqual(instance.cleanup_code(), 0)
        instance.stop()

    def test4(self):
        print("Testing GalactICsInterface generate_particles")
        number_of_particles_halo = 1000
        instance = GalactICsInterface(**default_options)
        self.assertEqual(instance.initialize_code(), 0)
        self.assertEqual(instance.set_output_path(instance.get_output_directory()), 0)
        self.assertEqual(instance.set_halo_number_of_particles(number_of_particles_halo), 0)
        self.assertEqual(instance.set_generate_bulge_flag(False), 0)
        self.assertEqual(instance.set_generate_disk_flag(False), 0)
        self.assertEqual(instance.set_order_of_multipole_expansion(0), 0)
        self.assertEqual(instance.commit_parameters(), 0)

        self.assertEqual(list(instance.get_number_of_particles_updated().values()), [0, 0])
        self.assertEqual(instance.generate_particles(), 0)
        self.assertEqual(list(instance.get_number_of_particles_updated().values()), [number_of_particles_halo, 0])

        masses, errors = instance.get_mass(range(number_of_particles_halo))
        self.assertEqual(errors, numpy.zeros(number_of_particles_halo))
        self.assertAlmostRelativeEquals(masses, numpy.ones(number_of_particles_halo)*masses[0])
        total_mass = masses.sum()

        if platform.processor() == 'ppc64le':
            # on ppc64le, the model generation has small differences from intel
            # change expected pos
            expected_mean_pos = numpy.array([73.5628, 76.251034, 75.53434])
        else:
            expected_mean_pos = numpy.array([73.768384103536604, 76.03533643054962, 75.176319462463255])

        expected_mean_vel = numpy.array([0.92904859858192501, 0.94953939936682585, 0.92897711758688095])

        x_positions, y_positions, z_positions, errors = instance.get_position(range(number_of_particles_halo))
        self.assertEqual(errors, numpy.zeros(number_of_particles_halo))
        self.assertAlmostEqual(numpy.array([numpy.mean(x_positions), numpy.mean(y_positions),
            numpy.mean(z_positions)]), numpy.array([0.0]*3), 5)
        self.assertAlmostRelativeEquals(numpy.array([numpy.mean(abs(x_positions)), numpy.mean(abs(y_positions)),
            numpy.mean(abs(z_positions))]), expected_mean_pos, 3)

        x_velocities, y_velocities, z_velocities, errors = instance.get_velocity(range(number_of_particles_halo))
        self.assertEqual(errors, numpy.zeros(number_of_particles_halo))
        self.assertAlmostEqual(numpy.array([numpy.mean(x_velocities), numpy.mean(y_velocities),
            numpy.mean(z_velocities)]), numpy.array([0.0]*3))
        self.assertAlmostRelativeEquals(numpy.array([numpy.mean(abs(x_velocities)), numpy.mean(abs(y_velocities)),
            numpy.mean(abs(z_velocities))]), expected_mean_vel, 2)

        self.assertEqual(instance.cleanup_code(), 0)
        instance.stop()


class GalactICsTests(TestWithMPI):

    default_unit_converter = nbody_system.nbody_to_si(1.0 | units.kpc, 1.0e6 | units.MSun)

    def test1(self):
        print("Testing GalactICs initialization")
        instance = GalactICs(**default_options)
        instance.initialize_code()
        instance.parameters.generate_bulge_flag = False
        instance.parameters.generate_disk_flag = False
        instance.parameters.order_of_multipole_expansion = 0
        instance.commit_parameters()
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        print("Testing GalactICs parameters (with unit converter)")
        instance = GalactICs(self.default_unit_converter, **default_options)
        instance.initialize_code()

        for par, value in [('generate_halo_flag', True), ('generate_disk_flag', True),
                ('generate_bulge_flag', True), ('halo_do_center_flag', True),
                ('bulge_do_center_flag', True), ('disk_do_center_flag', True)]:
            self.assertTrue(value is getattr(instance.parameters, par))
            setattr(instance.parameters, par, not value)
            self.assertFalse(value is getattr(instance.parameters, par))

        for par, value in [('number_of_grid_intervals', 90000), ('order_of_multipole_expansion', 10),
                ('number_of_radial_steps_correction_fns_disk_df', 10),
                ('number_of_iterations_disk_df', 50), ('halo_number_of_particles', 200000),
                ('bulge_number_of_particles', 50000), ('disk_number_of_particles', 100000),
                ('halo_random_seed', -1), ('bulge_random_seed', -1), ('disk_random_seed', -1)]:
            self.assertEqual(value, getattr(instance.parameters, par))
            setattr(instance.parameters, par, 1)
            self.assertEqual(1, getattr(instance.parameters, par))

        for par, value in [('halo_outer_radius', 300.0 | nbody_system.length),
                ('halo_scale_velocity', 3.26331115 | nbody_system.speed),
                ('halo_scale_radius', 6.06699419 | nbody_system.length),
                ('halo_truncation_width', 100.0 | nbody_system.length)]:
            self.assertEqual(instance.unit_converter.to_si(value),
                getattr(instance.parameters, par))
            setattr(instance.parameters, par, 3.0 | value.unit)
            self.assertEqual(instance.unit_converter.to_si(3.0 | value.unit),
                getattr(instance.parameters, par))

        self.assertEqual(os.path.join(instance.get_output_directory()), instance.parameters.output_directory)
        instance.parameters.output_directory = 'test'
        self.assertEqual("test", instance.parameters.output_directory)

        instance.cleanup_code()
        instance.stop()

    def test3(self):
        print("Testing GalactICs parameters (nbody units, no converter)")
        instance = GalactICs(**default_options)
        instance.initialize_code()

        for par, value in [('halo_outer_radius', 300.0 | nbody_system.length),
                ('halo_scale_velocity', 3.26331115 | nbody_system.speed),
                ('halo_scale_radius', 6.06699419 | nbody_system.length),
                ('halo_truncation_width', 100.0 | nbody_system.length)]:
            self.assertEqual(value, getattr(instance.parameters, par))
            setattr(instance.parameters, par, 3.0 | value.unit)
            self.assertEqual(3.0 | value.unit, getattr(instance.parameters, par))
        instance.cleanup_code()
        instance.stop()

    def slowtest4(self):
        print("Testing GalactICs generate_particles")
        n_particles_halo = 100
        n_particles_bulge = 100
        n_particles_disk = 100
        number_of_particles = n_particles_disk + n_particles_bulge + n_particles_halo

        instance = GalactICs(**default_options)
        instance.initialize_code()
        instance.parameters.disk_number_of_particles = n_particles_disk
        instance.parameters.bulge_number_of_particles = n_particles_bulge
        instance.parameters.halo_number_of_particles = n_particles_halo
        instance.commit_parameters()
        instance.generate_particles()
        self.assertEqual(len(instance.particles), number_of_particles)
        self.assertAlmostRelativeEquals(instance.particles.total_mass(), 1225.4466176 | nbody_system.mass, 3)
        self.assertAlmostRelativeEquals(instance.particles.kinetic_energy(), 2564.69894361 | nbody_system.energy, 3)
        self.assertAlmostRelativeEquals(instance.particles.potential_energy(G=nbody_system.G), -4531.58416742 | nbody_system.energy, 3)
        self.assertAlmostRelativeEquals(instance.particles.virial_radius(), 165.694750127 | nbody_system.length, 3)

        instance.cleanup_code()
        instance.stop()

    def test5(self):
        print("Testing GalactICs generate_particles")
        instance = GalactICs(**default_options)
        instance.initialize_code()
        instance.parameters.halo_number_of_particles = 1000
        instance.parameters.generate_bulge_flag = False
        instance.parameters.generate_disk_flag = False
        instance.parameters.order_of_multipole_expansion = 0
        instance.commit_parameters()
        instance.generate_particles()
        self.assertEqual(len(instance.particles), 1000)
        accuracy = 3
        mass_halo = 1178.89297009 | nbody_system.mass
        expected_kinetic_energy = 2418.49730735 | nbody_system.energy
        self.assertAlmostRelativeEquals(instance.particles.total_mass(), mass_halo, accuracy)
        self.assertAlmostRelativeEquals(instance.particles.kinetic_energy(), expected_kinetic_energy, accuracy)

        self.assertEqual(len(instance.halo_particles), 1000)
        self.assertEqual(len(instance.disk_particles), 0)
        self.assertEqual(len(instance.bulge_particles), 0)

        instance.cleanup_code()
        instance.stop()

    def test6(self):
        print("Testing GalactICs generate_particles: generate multiple sets")
        number_of_particles = 1000
        instance = GalactICs(**default_options)
        instance.initialize_code()
        instance.parameters.halo_number_of_particles = number_of_particles
        instance.parameters.generate_bulge_flag = False
        instance.parameters.generate_disk_flag = False
        instance.parameters.order_of_multipole_expansion = 0
        instance.parameters.halo_random_seed = -1.0
        instance.commit_parameters()

        instance.generate_particles()
        set1 = instance.particles.copy()
        self.assertEqual(len(set1), number_of_particles)

        instance.generate_particles()
        set2 = instance.particles.copy()
        self.assertEqual(len(set2), number_of_particles)
        # GalactICs' random-number generator is re-seeded with 'halo_random_seed'
        # each time, and the result should be the same:
        for attribute in ["mass", "x", "y", "z", "vx", "vy", "vz"]:
            self.assertEqual(getattr(set1, attribute), getattr(set2, attribute))

        instance.parameters.halo_random_seed = -42.0
        instance.generate_particles()
        # halo_random_seed changed: draw a different random set of particles
        set3 = instance.particles.copy()
        self.assertEqual(len(set3), number_of_particles)
        self.assertEqual(set1.mass, set3.mass)
        self.assertRaises(self.failureException, self.assertEqual, set1.x, set3.x)
        self.assertAlmostRelativeEquals(abs(set1.x).median(), abs(set3.x).median(), 1)
        self.assertAlmostRelativeEquals(abs(set1.vy).median(), abs(set3.vy).median(), 1)

        instance.cleanup_code()
        instance.stop()

    def test7(self):
        print("Testing GalactICs state")
        number_of_particles = 1000

        print("First do everything manually:")
        instance = GalactICs(**default_options)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.parameters.halo_number_of_particles = number_of_particles
        instance.parameters.generate_bulge_flag = False
        instance.parameters.generate_disk_flag = False
        instance.parameters.order_of_multipole_expansion = 0
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.overridden().generate_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'UPDATE')
        instance.invoke_state_change_updated()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        self.assertEqual(len(instance.particles), number_of_particles)
        instance.cleanup_code()
        self.assertEqual(instance.get_name_of_current_state(), 'END')
        instance.stop()

        print("initialize_code(), (re)commit_parameters(), update_particle_set(), "
            "and cleanup_code() should be called automatically:")
        instance = GalactICs(**default_options)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.parameters.halo_number_of_particles = number_of_particles
        instance.parameters.generate_bulge_flag = False
        instance.parameters.generate_disk_flag = False
        instance.parameters.order_of_multipole_expansion = 0
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        self.assertEqual(instance.get_number_of_particles_updated(), 0)
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.parameters.halo_random_seed = -42.0
        self.assertEqual(instance.get_name_of_current_state(), 'CHANGE_PARAMETERS_EDIT')
        self.assertEqual(instance.get_number_of_particles_updated(), 0)
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.generate_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        self.assertEqual(len(instance.particles), number_of_particles)
        self.assertEqual(instance.get_number_of_particles_updated(), 0)
        instance.stop()
        self.assertEqual(instance.get_name_of_current_state(), 'STOPPED')



import os.path
import numpy
from amuse.test.amusetest import TestWithMPI

from amuse.community.krome.interface import KromeInterface, Krome, solar_abundances
from amuse.units import units
from amuse.datamodel import Particles

from amuse.io import read_set_from_file

# default_options={}
default_options = dict(redirection="none")
# default_options=dict(debugger="gdb")


class TestKromeInterface(TestWithMPI):

    def test1(self):
        print("Test 1: initialization")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test2(self):
        print("Test 1: add particle, get state")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = 1.e5
        t = 500.
        ion = 1.e-11
        id, err = instance.new_particle(dens, t, ion)

        self.assertEqual(err, 0)

        self.assertEqual(instance.commit_particles(), 0)

        dens_, t_, ion_, err = instance.get_state(id)

        self.assertEqual(err, 0)

        self.assertEqual(dens_, dens)
        self.assertEqual(t_, t)
        self.assertEqual(ion_, ion)

        self.assertEqual(0, instance.cleanup_code())

        instance.stop()

    def test3(self):
        print("Test 1: add 2 particles, get state")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = [1.e5, 2.e5]
        t = [500., 550]
        ion = [1.e-11, 2.e-11]
        id, err = instance.new_particle(dens, t, ion)

        self.assertEqual(err, 0)

        self.assertEqual(instance.commit_particles(), 0)

        for i in range(2):
            dens_, t_, ion_, err = instance.get_state(id[i])

            self.assertEqual(err, 0)

            self.assertEqual(dens_, dens[i])
            self.assertEqual(t_, t[i])
            self.assertEqual(ion_, ion[i])

        self.assertEqual(0, instance.cleanup_code())

        instance.stop()

    def test4(self):
        print("Test 1: add 100 particles, get state")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = 1.e5*numpy.random.random(100)
        t = 500.*numpy.random.random(100)
        ion = 1.e-11*numpy.random.random(100)
        id, err = instance.new_particle(dens, t, ion)

        self.assertEqual(err, 0)

        self.assertEqual(instance.commit_particles(), 0)

        for i in range(100):
            dens_, t_, ion_, err = instance.get_state(id[i])

            self.assertEqual(err, 0)

            self.assertEqual(dens_, dens[i])
            self.assertEqual(t_, t[i])
            self.assertEqual(ion_, ion[i])

        self.assertEqual(0, instance.cleanup_code())

        instance.stop()

    def test5(self):
        print("Test 5: can we get species?")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)

        first, last, err = instance.get_firstlast_abundance()
        self.assertEqual(err, 0)
        self.assertTrue(last-first > 0)

        for i in range(first, last+1):
            name, err = instance.get_name_of_species(i)
            print(name)
            self.assertEqual(err, 0)
            index, err = instance.get_index_of_species(name)
            self.assertEqual(i, index)

        instance.stop()

    def test6(self):
        print("Test 6: add 100 particles, remove particles")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = 1.e5*numpy.random.random(100)
        t = 500.*numpy.random.random(100)
        ion = 1.e-11*numpy.random.random(100)
        ids, err = instance.new_particle(dens, t, ion)

        self.assertEqual(err, 0)

        self.assertEqual(instance.commit_particles(), 0)

        for i in ids[:10]:
            instance.delete_particle(i)

        instance.recommit_particles()

        for i in range(10, 100):
            dens_, t_, ion_, err = instance.get_state(ids[i])

            self.assertEqual(err, 0)

            self.assertEqual(dens_, dens[i])
            self.assertEqual(t_, t[i])
            self.assertEqual(ion_, ion[i])

        dens_, t_, ion_, err = instance.get_state(ids[0])

        self.assertEqual(err, -1)

        self.assertEqual(0, instance.cleanup_code())

        instance.stop()

    def test7(self):
        print("Test 1: add particle, set abundances")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = 1.e5
        t = 500.
        ion = 1.e-11
        id, err = instance.new_particle(dens, t, ion)

        instance.commit_particles()

        first, last, err = instance.get_firstlast_abundance()
        for i in range(first, last+1):
            x, err = instance.get_abundance(id, i)
            self.assertTrue((x >= 0.) & (x <= 1.))
            self.assertEqual(err, 0)
        x, err = instance.get_abundance(id, last+1)
        self.assertEqual(err, -1)

        for s in ["H", "HE", "C", "SI", "O"]:
            x = solar_abundances[s]
            aid, err = instance.get_index_of_species(s)
            instance.set_abundance(id, aid, x)

        for s in ["H", "HE", "C", "SI", "O"]:
            x = solar_abundances[s]
            aid, err = instance.get_index_of_species(s)
            xx, err = instance.get_abundance(id, aid)
            self.assertEqual(x, xx)
            self.assertEqual(err, 0)

    def test8(self):
        print("evolve test")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = 1.e2
        t = 100.
        ion = 2.e-17
        id, err = instance.new_particle(dens, t, ion)

        first, last, err = instance.get_firstlast_abundance()
        for i in range(first, last+1):
            err = instance.set_abundance(id, i, 1.e-40)

        for s in ["H", "HE", "C", "SI", "O"]:
            x = solar_abundances[s]
            aid, err = instance.get_index_of_species(s)
            instance.set_abundance(id, aid, x)

        instance.commit_particles()

        yr = 365*24*3600.
        err = instance.evolve_model(10000.*yr)
        self.assertEqual(err, 0)
        time, err = instance.get_time()
        self.assertEqual(err, 0)
        self.assertEqual(time, 10000.*yr)

        first, last, err = instance.get_firstlast_abundance()
        for i in range(first, last+1):
            x, err = instance.get_abundance(id, i)
            self.assertEqual(err, 0)
            name, err = instance.get_name_of_species(i)
            print(i, name, x)

    def test9(self):
        print("evolve test 2")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = 1.e2
        t = 100.
        ion = 2.e-17
        id, err = instance.new_particle(dens, t, ion)

        first, last, err = instance.get_firstlast_abundance()
        for i in range(first, last+1):
            err = instance.set_abundance(id, i, 1.e-40)

        for s in ["H", "HE", "C", "SI", "O"]:
            x = solar_abundances[s]
            aid, err = instance.get_index_of_species(s)
            instance.set_abundance(id, aid, x)

        aid, err = instance.get_index_of_species("C")
        instance.set_abundance(id, aid, 1.e-40)

        aid, err = instance.get_index_of_species("C+")
        instance.set_abundance(id, aid, solar_abundances["C"])

        aid, err = instance.get_index_of_species("H2")
        instance.set_abundance(id, aid, 1.e-6)
        aid, err = instance.get_index_of_species("H+")
        instance.set_abundance(id, aid, 1.e-4)

        instance.commit_particles()

        yr = 365*24*3600.
        err = instance.evolve_model(10000.*yr)
        self.assertEqual(err, 0)
        time, err = instance.get_time()
        self.assertEqual(err, 0)
        self.assertEqual(time, 10000.*yr)

        first, last, err = instance.get_firstlast_abundance()
        for i in range(first, last+1):
            x, err = instance.get_abundance(id, i)
            self.assertEqual(err, 0)
            name, err = instance.get_name_of_species(i)
            print(i, name, x)

    def test10(self):
        print("check initialization of abundances")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = 1.e2
        t = 100.
        ion = 2.e-17
        id, err = instance.new_particle(dens, t, ion)

        abundances = {"E": 0.000369180975425,
                     "H+": 0.0001, "HE": 0.0775,
                     "C+": 0.000269180975425, "SI": 3.2362683404e-05,
                     "O": 0.000489828841345}

        first, last, err = instance.get_firstlast_abundance()
        for i in range(first, last+1):
            x, err = instance.get_abundance(id, i)
            self.assertEqual(err, 0)
            name, err = instance.get_name_of_species(i)
            if name in abundances:
                self.assertAlmostEqual(x, abundances[name], 12)

    def test11(self):
        print("evolve test, comparison")

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = 1.e2
        t = 100.
        ion = 2.e-17
        id, err = instance.new_particle(dens, t, ion)
        instance.evolve_model(1.e10)

        result1 = {}
        first, last, err = instance.get_firstlast_abundance()
        for i in range(first, last+1):
            x, err = instance.get_abundance(id, i)
            self.assertEqual(err, 0)
            name, err = instance.get_name_of_species(i)
            result1[name] = x

        instance = self.new_instance_of_an_optional_code(KromeInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        dens = 1.e2
        t = 100.
        ion = 2.e-17
        id, err = instance.new_particle(dens, t, ion)
        instance.evolve_model(1.e9)
        instance.evolve_model(5.e9)
        instance.evolve_model(1.e10)

        result2 = {}
        first, last, err = instance.get_firstlast_abundance()
        for i in range(first, last+1):
            x, err = instance.get_abundance(id, i)
            self.assertEqual(err, 0)
            name, err = instance.get_name_of_species(i)
            result2[name] = x

        for x in result1:
            self.assertAlmostEqual(result1[x], result2[x])


class TestKrome(TestWithMPI):
    def makeparts(self, N):
        parts = Particles(N)
        numpy.random.seed(1234567)
        parts.number_density = (numpy.random.random(N)*1.e5+1.e5) | units.cm**-3
        parts.temperature = (numpy.random.random(N)*500+100) | units.K
        parts.ionrate = (numpy.random.random(N)*1.e-11+1.e-17) | units.s**-1
        return parts

    def test0(self):
        print("test1: basic startup and flow")
        instance = self.new_instance_of_an_optional_code(Krome)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.commit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')

        instance.cleanup_code()
        instance.stop()

    def test1(self):
        print("test1: adding particles")

        instance = self.new_instance_of_an_optional_code(Krome)

        parts = self.makeparts(5)

        self.assertEqual(len(instance.particles), 0)
        instance.particles.add_particles(parts)
        self.assertEqual(len(instance.particles), len(parts))

        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')

        part2 = instance.particles.copy()

        self.assertAlmostRelativeEquals(parts.number_density, part2.number_density, 12)
        self.assertAlmostRelativeEquals(parts.temperature, part2.temperature, 12)
        self.assertAlmostRelativeEquals(parts.ionrate, part2.ionrate, 12)

        for p in part2:
            i = instance.species["E"]
            self.assertAlmostEqual(p.abundances[i], 0.000369180975425)
            i = instance.species["H+"]
            self.assertAlmostEqual(p.abundances[i], 0.0001)
            i = instance.species["HE"]
            self.assertAlmostEqual(p.abundances[i], 0.0775)
            i = instance.species["C+"]
            self.assertAlmostEqual(p.abundances[i], 0.000269180975425)
            i = instance.species["SI"]
            self.assertAlmostEqual(p.abundances[i], 3.2362683404e-05)
            i = instance.species["O"]
            self.assertAlmostEqual(p.abundances[i], 0.000489828841345)

        instance.cleanup_code()
        instance.stop()

    def test2(self):
        print("test2: adding particles w abund.")

        instance = self.new_instance_of_an_optional_code(Krome)

        parts = self.makeparts(5)

        N = len(instance.species)

        parts.abundances = numpy.zeros((5, N))

        for i in range(5):
            parts[i].abundances = (numpy.array(range(N))+1)/(N+1.)

        instance.particles.add_particles(parts)

        channel = parts.new_channel_to(instance.particles)
        channel.copy()

        part2 = instance.particles.copy()

        self.assertAlmostRelativeEquals(parts.number_density, part2.number_density, 12)
        self.assertAlmostRelativeEquals(parts.temperature, part2.temperature, 12)
        self.assertAlmostRelativeEquals(parts.ionrate, part2.ionrate, 12)

        for i in range(5):
            self.assertAlmostRelativeEquals(part2[i].abundances, parts[i].abundances, 12)

        instance.cleanup_code()
        instance.stop()

    def test3(self):
        print("test3: evolve test")

        instance = self.new_instance_of_an_optional_code(Krome, **default_options)

        parts = Particles(1)
        parts.number_density = 1.e5 | units.cm**-3
        parts.temperature = 50 | units.K
        parts.ionrate = 2.e-17 | units.s**-1

        Ns = len(instance.species)

        parts.abundances = numpy.zeros((1, Ns))

        instance.particles.add_particles(parts)

        instance.evolve_model(1. | units.Myr)

        print(instance.particles.abundances)

        f = 2*instance.particles[0].abundances[instance.species["H2"]]
        self.assertTrue(f > 0.95)  # not much of a test..
        # ~ for x,i in instance.species.items():
            # ~ print x, instance.particles[0].abundances[i]

        instance.cleanup_code()
        instance.stop()

    def test4(self):
        print("test4: evolve test (10 part)")

        instance = self.new_instance_of_an_optional_code(Krome, **default_options)

        parts = Particles(10)
        parts.number_density = 1.e5 | units.cm**-3
        parts.temperature = 50 | units.K
        parts.ionrate = 2.e-17 | units.s**-1

        Ns = len(instance.species)

        parts.abundances = numpy.zeros((10, Ns))

        instance.particles.add_particles(parts)

        instance.evolve_model(1. | units.Myr)

        f = 2*instance.particles[0].abundances[instance.species["H2"]]
        self.assertTrue(f > 0.95)  # not much of a test..
        # ~ for x,i in instance.species.items():
            # ~ print x, instance.particles[0].abundances[i]

        instance.cleanup_code()
        instance.stop()



from amuse.support.interface import InCodeComponentImplementation

from amuse.test.amusetest import TestWithMPI
from amuse.support import exceptions
from amuse.support import options

import os
import time
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.rfi.tools import create_c
from amuse.rfi import channel
from amuse.rfi.core import *

from . import test_c_implementation


class TestCSocketsImplementationInterface(test_c_implementation.TestCImplementationInterface):

    @classmethod
    def setup_class(cls):
        cls.check_not_in_mpiexec()
        super(TestCSocketsImplementationInterface, cls).setup_class()
        # set sockets channel as default channel
        options.GlobalOptions.instance().override_value_for_option("channel_type", "sockets")

    @classmethod
    def teardown_class(cls):
        del options.GlobalOptions.instance().overriden_options["channel_type"]

    def test22(self):
        self.skip("this test uses mpi internals, skip here")

    def test29(self):
        self.skip("this test uses mpi internals, skip here")

    @classmethod
    def check_not_in_mpiexec(cls):
        """
        The tests will fork another process, if the test run 
        is itself an mpi process, the tests may fail.

        For the hydra process manager the tests will fail.
        So skip the tests if we detect hydra
        """
        if 'HYDI_CONTROL_FD' in os.environ:
            return  # for now assume HYDI_CONTROL_FD is newer, and sockets will work!
        if 'HYDRA_CONTROL_FD' in os.environ or 'PMI_FD' in os.environ:
            cls.skip('cannot run the socket tests under mpi process manager')



import numpy
import time
import sys
import pickle

from amuse.test import amusetest
from amuse.units import units
from amuse.units import constants
from amuse.units import nbody_system
from amuse.support.exceptions import AmuseException
from amuse.support.interface import InCodeComponentImplementation

from amuse import datamodel


class TestParticlesProperties(amusetest.TestCase):

    def test1(self):

        particles = datamodel.Particles(2)
        particles.mass = 10 | units.kg

        self.assertTrue(hasattr(particles, 'collection_attributes'))

        particles.collection_attributes.timestamp = 1 | units.yr
        self.assertEqual(particles.collection_attributes.timestamp,  1 | units.yr)

        particles.collection_attributes.a = 2
        self.assertEqual(particles.collection_attributes.a,  2)

    def test2(self):

        particles = datamodel.Particles(2)
        particles.collection_attributes.timestamp = 1 | units.yr

        self.assertEqual(str(particles.collection_attributes), "timestamp: 1 yr")

        particles.collection_attributes.a = 2
        self.assertEqual(str(particles.collection_attributes), "timestamp: 1 yr\na: 2")

    def test3(self):

        particles1 = datamodel.Particles(2)
        particles1.collection_attributes.timestamp = 1 | units.yr
        particles1.collection_attributes.a = 2
        particles2 = particles1.copy()

        self.assertEqual(particles2.collection_attributes.timestamp,  1 | units.yr)
        self.assertEqual(particles2.collection_attributes.a,  2)
        self.assertEqual(str(particles2.collection_attributes), "timestamp: 1 yr\na: 2")

    def test4(self):

        particles1 = datamodel.Particles(2)
        particles1.collection_attributes.timestamp = 1 | units.yr
        particles1.collection_attributes.a = 2
        pickled_string = pickle.dumps(particles1)
        particles2 = pickle.loads(pickled_string)
        self.assertEqual(particles2.collection_attributes.timestamp,  1 | units.yr)
        self.assertEqual(particles2.collection_attributes.a,  2)



import os
import sys
import numpy

from amuse.community.octgrav.interface import OctgravInterface, Octgrav

from amuse.test.amusetest import TestWithMPI
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.rfi import channel
from amuse.ic.plummer import *


class TestMPIInterface(TestWithMPI):

    def test1(self):

        instance = self.new_instance_of_an_optional_code(OctgravInterface)
        instance.new_particle(11.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.0)
        retrieved_state = instance.get_state(1)
        self.assertEqual(11.0,  retrieved_state['mass'])
        self.assertEqual(2.0, retrieved_state['radius'])
        self.assertEqual(instance.get_number_of_particles()['number_of_particles'], 1)
        instance.cleanup_code()
        instance.stop()

    def test2(self):

        instance = self.new_instance_of_an_optional_code(OctgravInterface)
        instance.initialize_code()
        instance.new_particle(
            [1, 10, 100],
            [3, 12, 102],
            [4, 13, 103],
            [5, 14, 104],
            [6, 15, 105],
            [7, 16, 106],
            [8, 17, 107],
            [2, 11, 101])

        particle1_state = instance.get_state(1)
        self.assertEqual(1,   particle1_state['mass'])

        particle2_state = instance.get_state(2)
        self.assertEqual(10,  particle2_state['mass'])

        instance.delete_particle(1)

        size_result = instance.get_number_of_particles()
        self.assertEqual(2, size_result['number_of_particles'])

        new_particle1_state = instance.get_state(1)
        self.assertEqual(10,  new_particle1_state['mass'])

        new_particle_result = instance.new_particle(
                                  1000,
                                  1002,
                                  1003,
                                  1004,
                                  1005,
                                  1006,
                                  1007,
                                  1001)
        self.assertEqual(4, new_particle_result['index_of_the_particle'], 4)

        new_particle4_state = instance.get_state(4)
        self.assertEqual(1000,  new_particle4_state['mass'])

        instance.cleanup_code()
        instance.stop()

    def test3(self):
        instance = self.new_instance_of_an_optional_code(OctgravInterface)
        instance.initialize_code()
        instance.set_eps2(0.1**2)
        instance.commit_parameters()
        ids = []
        for i in range(32):
            id, errorcode = instance.new_particle(mass=10.0, radius=1.0, x=i, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
            ids.append(id)
            self.assertEqual(errorcode, 0)
        instance.commit_particles()
        potential, errorcode = instance.get_potential(ids[0])
        self.assertEqual(errorcode, 0)
        excpected_potential = numpy.sum([-10.0 / numpy.sqrt((x+1.0)**2 + 0.1**2) for x in range(31)])
        self.assertAlmostRelativeEquals(potential, excpected_potential, 5)
        total_potential, errorcode = instance.get_potential_energy()
        potentials, errorcode = instance.get_potential(ids)

        self.assertAlmostRelativeEquals(total_potential, numpy.sum(potentials * 10.0) / 2.0)


class TestAmuseInterface(TestWithMPI):
    def new_system_of_sun_and_earth(self):
        stars = datamodel.Stars(2)
        sun = stars[0]
        sun.mass = units.MSun(1.0)
        sun.position = units.m(numpy.array((0.0, 0.0, 0.0)))
        sun.velocity = units.ms(numpy.array((0.0, 0.0, 0.0)))
        sun.radius = units.RSun(1.0)

        earth = stars[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = units.km(numpy.array((149.5e6, 0.0, 0.0)))
        earth.velocity = units.ms(numpy.array((0.0, 29800, 0.0)))

        return stars

    def test0(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, units.AU)
        instance = self.new_instance_of_an_optional_code(Octgrav, convert_nbody)
        instance.initialize_code()
        self.assertAlmostRelativeEqual(0.01, instance.parameters.epsilon_squared.value_in(units.AU**2), 2)  # default
        instance.parameters.epsilon_squared = 0.05 | units.AU**2
        self.assertAlmostRelativeEqual(0.05, instance.parameters.epsilon_squared.value_in(units.AU**2), 6)

        self.assertAlmostEqual(0.8, instance.parameters.opening_angle, 6)  # default
        instance.parameters.opening_angle = 0.5
        self.assertAlmostEqual(0.5, instance.parameters.opening_angle, 6)
        instance.parameters.timestep = 1.0 | units.s
        self.assertEqual(1.0 | units.s, instance.parameters.timestep)
        instance.stop()

    def test1(self):
        plummer_size = 500
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)
        stars = new_plummer_model(plummer_size, convert_nbody)
        stars.radius = range(1, plummer_size+1) | units.km

        instance = self.new_instance_of_an_optional_code(Octgrav, convert_nbody)
        instance.particles.add_particles(stars)

        instance.evolve_model(5 | units.day)
        energy_total_init = instance.potential_energy + instance.kinetic_energy
        instance.evolve_model(100 | units.day)
        energy_total_final = instance.potential_energy + instance.kinetic_energy

        self.assertAlmostRelativeEqual(energy_total_init, energy_total_final, 2)
        instance.stop()

    def test2(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        instance = self.new_instance_of_an_optional_code(Octgrav, convert_nbody)
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.0 | units.AU**2
        instance.parameters.stopping_conditions_number_of_steps = 1
        self.assertEqual(instance.parameters.stopping_conditions_number_of_steps, 1)

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]

        instance.particles.add_particles(stars)
        instance.stopping_conditions.number_of_steps_detection.enable()
        instance.evolve_model(365.0 | units.day)
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        instance.cleanup_code()
        instance.stop()

    def test3(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.radius = 0.005 | nbody_system.length
        particles.vx = 0 | nbody_system.speed
        particles.vy = 0 | nbody_system.speed
        particles.vz = 0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        instance = self.new_instance_of_an_optional_code(Octgrav)
        instance.initialize_code()
        instance.parameters.stopping_conditions_number_of_steps = 20
        self.assertEqual(instance.parameters.stopping_conditions_number_of_steps, 20)
        instance.parameters.epsilon_squared = (0.01 | nbody_system.length)**2
        instance.particles.add_particles(particles)
        instance.stopping_conditions.number_of_steps_detection.enable()
        instance.evolve_model(10 | nbody_system.time)
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        self.assertTrue(instance.model_time < 10 | nbody_system.time)
        instance.stop()

    def test4(self):
        plummer_size = 500
        stars = new_plummer_model(plummer_size)
        stars.radius = 0 | nbody_system.length

        instance = self.new_instance_of_an_optional_code(Octgrav)
        instance.particles.add_particles(stars)

        instance.synchronize_model()

        ax, ay, az = instance.get_gravity_at_point(0. | nbody_system.length,
                0. | nbody_system.length,
                100. | nbody_system.length,
                0. | nbody_system.length)

        self.assertAlmostEqual(ax.number, 0., 3)
        self.assertAlmostRelativeEqual(ay.number, -1./100**2, 3)
        self.assertAlmostEqual(az.number, 0., 3)

        pot = instance.get_potential_at_point([0., 0.] | nbody_system.length,
                [0., 100] | nbody_system.length,
                [100., 0.] | nbody_system.length,
                [0., 0.] | nbody_system.length)

        self.assertAlmostRelativeEqual(pot.number, [-1/100., -1/100.], 3)

        instance.stop()



from amuse.test import amusetest

from amuse.units import constants
from amuse.units.generic_unit_converter import *


class TestGenericUnits(amusetest.TestCase):
    def test1(self):
        L = 1 | length
        T = 1 | time
        V = 1 | speed
        self.assertTrue(L/T == V)

    def test2(self):
        # natural units
        convert_generic = ConvertBetweenGenericAndSiUnits(constants.c, constants.hbar, constants.G, constants.kB)
        M = 1 | mass
        T = 1 | time
        L = 1 | length
        H = 1 | temperature

        M_in_si = convert_generic.to_si(M)
        T_in_si = convert_generic.to_si(T)
        L_in_si = convert_generic.to_si(L)
        H_in_si = convert_generic.to_si(H)

        self.assertAlmostRelativeEqual(M_in_si, 2.1764411e-8 | units.kg, 3)
        self.assertAlmostRelativeEqual(T_in_si, 5.39124e-44 | units.s, 3)
        self.assertAlmostRelativeEqual(L_in_si, 1.616252e-35 | units.m, 3)
        self.assertAlmostRelativeEqual(H_in_si, 1.416785e32 | units.K, 3)

    def test3(self):
        # Gadget units
        UnitLength_in_cm = 3.085678e21 | units.cm  # 1.0 kpc
        UnitMass_in_g = 1.989e43 | units.g    # 1.0e10 solar masses
        UnitVelocity_in_cm_per_s = 1e5 | units.cm / units.s  # 1 km/sec

        convert_generic = ConvertBetweenGenericAndSiUnits(UnitLength_in_cm, UnitMass_in_g, UnitVelocity_in_cm_per_s)
        M = 1 | mass
        T = 1 | time
        L = 1 | length

        M_in_si = convert_generic.to_si(M)
        T_in_si = convert_generic.to_si(T)
        L_in_si = convert_generic.to_si(L)

        self.assertAlmostRelativeEqual(M_in_si, 1.989e40 | units.kg, 3)
        self.assertAlmostRelativeEqual(T_in_si, 3.085678e16 | units.s, 3)
        self.assertAlmostRelativeEqual(L_in_si, 3.085678e19 | units.m, 3)

    def test4(self):
        print("Generic units and vector quantities")
        UnitLength_in_cm = 3.085678e21 | units.cm  # 1.0 kpc
        UnitMass_in_g = 1.989e43 | units.g    # 1.0e10 solar masses
        UnitVelocity_in_cm_per_s = 1e5 | units.cm / units.s  # 1 km/sec

        convert_generic = ConvertBetweenGenericAndSiUnits(UnitLength_in_cm, UnitMass_in_g, UnitVelocity_in_cm_per_s)
        M = [1, 2] | mass
        T = [1, 2] | time
        L = [1, 2] | length

        M_in_si = convert_generic.to_si(M)
        T_in_si = convert_generic.to_si(T)
        L_in_si = convert_generic.to_si(L)

        self.assertAlmostEqual(M_in_si, [1.989e40, 2*1.989e40] | units.kg, 3, in_units=1.0e10*units.MSun)
        self.assertAlmostEqual(T_in_si, [3.085678e16, 2*3.085678e16] | units.s, 3, in_units=units.s*units.kpc/units.km)
        self.assertAlmostEqual(L_in_si, [3.085678e19, 2*3.085678e19] | units.m, 3, in_units=3.085678e19*units.m)

    def test5(self):

        self.assertRaises(UnitsNotOrtogonalException, ConvertBetweenGenericAndSiUnits, 1 | units.km, 2 | units.s,
            3 | units.km / units.s, expected_message="The number of orthoganal units is incorrect, expected 3 "
            "but found 2. To convert between S.I. units and another system of units a set of quantities with "
            "orthogonal units is needed. These can be quantities with a single unit (such as length or time) "
            "or quantities with a derived units (such as velocity or force)")

    def test6(self):
        self.assertRaises(NotAQuantityException, ConvertBetweenGenericAndSiUnits, 1)
        self.assertRaises(NotAScalarException, ConvertBetweenGenericAndSiUnits,  1 | units.km, [1, 2] | units.s, )

    def test7(self):
        converter = ConvertBetweenGenericAndSiUnits(1.0 | units.MSun, 1.0 | units.AU, 1.0 | units.yr)
        G_in_generic = converter.to_generic(constants.G)
        self.assertEqual(str(G_in_generic.unit), 'length**3 * mass**-1 * time**-2')
        self.assertAlmostEqual(G_in_generic.number, 39.4852492465)

        time_unit = ((1.0 | units.AU)**3 / (constants.G * (1.0 | units.MSun))).sqrt().as_quantity_in(units.yr)
        self.assertAlmostEqual(time_unit, 0.159141174195 | units.yr)
        converter = ConvertBetweenGenericAndSiUnits(1.0 | units.MSun, 1.0 | units.AU, time_unit)
        G_in_generic = converter.to_generic(constants.G)
        self.assertEqual(str(G_in_generic.unit), 'length**3 * mass**-1 * time**-2')
        self.assertAlmostEqual(G_in_generic.number, 1.0)



from amuse.support.interface import InCodeComponentImplementation

from amuse.test.amusetest import TestWithMPI, TestCase

import numpy
import sys
import os
import time
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.rfi import python_code
from amuse.rfi.core import *
from amuse.rfi.async_request import AsyncRequestsPool

from . import test_python_implementation
from . import test_python_implementation_mpi


class TestInterfaceSockets(test_python_implementation.TestInterface):
    def setUp(self):
        self.check_not_in_mpiexec()
        super(test_python_implementation.TestInterface, self).setUp()

    def check_not_in_mpiexec(self):
        """
        The tests will fork another process, if the test run
        is itself an mpi process, the tests may fail. 

        For the hydra process manager the tests will fail.
        So skip the tests if we detect hydra
        """
        if 'HYDRA_CONTROL_FD' in os.environ or 'PMI_FD' in os.environ:
            self.skip('cannot run the socket tests under hydra process manager')

    def ForTesting(self, **options):
        options["worker_dir"] = self.get_path_to_results()
        options["channel_type"] = "sockets"
        return test_python_implementation.ForTesting(**options)

    def ForTestingInterface(self, **options):
        options["worker_dir"] = self.get_path_to_results()
        options["channel_type"] = "sockets"
        return test_python_implementation.ForTestingInterface(**options)

    def test2(self):
        implementation = test_python_implementation.ForTestingImplementation()
        x = python_code.PythonImplementation(implementation, test_python_implementation.ForTestingInterface)

        input_message = python_code.SocketMessage(0, 10, 1)
        input_message.ints = [1]

        output_message = python_code.SocketMessage(0, 10, 1)

        x.handle_message(input_message, output_message)

        self.assertEqual(len(output_message.ints), 1)
        self.assertEqual(len(output_message.doubles), 1)
        self.assertEqual(output_message.ints[0], 0)
        self.assertEqual(output_message.doubles[0], 0.0)

    def test3(self):
        implementation = test_python_implementation.ForTestingImplementation()
        x = python_code.PythonImplementation(implementation, test_python_implementation.ForTestingInterface)

        input_message = python_code.SocketMessage(0, 11, 1)
        input_message.ints = [1]
        input_message.doubles = [12.0]

        output_message = python_code.SocketMessage(0, 10, 1)

        x.handle_message(input_message, output_message)

        self.assertEqual(len(output_message.ints), 1)
        self.assertEqual(len(output_message.doubles), 0)
        self.assertEqual(output_message.ints[0], 0)
        self.assertEqual(implementation.masses[1], 12.0)

    def test4(self):
        implementation = test_python_implementation.ForTestingImplementation()
        x = python_code.PythonImplementation(implementation, test_python_implementation.ForTestingInterface)

        input_message = python_code.SocketMessage(0, 11, 4)
        input_message.ints = [1, 2, 3, 4]
        input_message.doubles = [12.0, 13.0, 14.0, 15.0]

        output_message = python_code.SocketMessage(0, 10, 4)

        x.handle_message(input_message, output_message)

        self.assertEqual(len(output_message.ints), 4)
        self.assertEqual(len(output_message.doubles), 0)
        self.assertEqual(output_message.ints[0], 0)
        self.assertEqual(output_message.ints[3], 0)
        self.assertEqual(implementation.masses[1], 12.0)
        self.assertEqual(implementation.masses[2], 13.0)
        self.assertEqual(implementation.masses[3], 14.0)
        self.assertEqual(implementation.masses[4], 15.0)

    def test8(self):
        implementation = test_python_implementation.ForTestingImplementation()
        x = python_code.PythonImplementation(implementation, test_python_implementation.ForTestingInterface)

        input_message = python_code.SocketMessage(0, 12, 1)
        input_message.ints = [20]

        output_message = python_code.SocketMessage(0, 10, 1)

        x.handle_message(input_message, output_message)

        self.assertEqual(len(output_message.ints), 2)
        self.assertEqual(output_message.ints[0], 0)
        self.assertEqual(output_message.ints[1], 20)

    def test27(self):
        pass  # skip because only supported for mpi channel


class TestInterfaceSocketsMPI(test_python_implementation_mpi.TestInterface):
    def setUp(self):
        self.check_not_in_mpiexec()
        super(test_python_implementation_mpi.TestInterface, self).setUp()

    def check_not_in_mpiexec(self):
        """
        The tests will fork another process, if the test run
        is itself an mpi process, the tests may fail. 

        For the hydra process manager the tests will fail.
        So skip the tests if we detect hydra
        """
        if 'HYDRA_CONTROL_FD' in os.environ or 'PMI_FD' in os.environ:
            self.skip('cannot run the socket tests under hydra process manager')

    def ForTesting(self, **options):
        options["worker_dir"] = self.get_path_to_results()
        options["channel_type"] = "sockets"
        return test_python_implementation_mpi.ForTesting(**options)

    def ForTestingInterface(self, **options):
        options["worker_dir"] = self.get_path_to_results()
        options["channel_type"] = "sockets"
        return test_python_implementation_mpi.ForTestingInterface(**options)



import os
import os.path
import numpy

from amuse.units import units
from amuse.datamodel import Particles, Particle
from amuse.support.exceptions import AmuseException
from amuse.test.amusetest import TestCase, get_path_to_results

from amuse.community.mesa.interface import MESA
from amuse.community.evtwin.interface import EVtwin
from amuse.community.sse.interface import SSE

from amuse.couple.fallback_stellar_evolution import FallbackStellarEvolution


class TestFallbackStellarEvolution(TestCase):

    def test1(self):
        print("Testing FallbackStellarEvolution")
        instance = FallbackStellarEvolution()
        instance.stop()

    def xtest2(self):
        print("Testing FallbackStellarEvolution: evolve tests")


# results of original code  (not really check the numbers, tests have been relaxed because
# different timestepping of evolve_model and evolve_one_step)
        results = dict()
        results["10.0 MSun"] = dict(sse_age=3644655.52487 | units.yr,
                                      sse_mass=9.99289747724 | units.MSun,
                                      sse_rad=4.24764389443 | units.RSun,
                                      sse_L=5993.72678228 | units.LSun,
                                      evtwin_age=2999999.99732 | units.yr,
                                      evtwin_mass=9.99832250931 | units.MSun,
                                      evtwin_rad=4.1727904434 | units.RSun,
                                      evtwin_L=6175.5214329 | units.LSun)
        results["20.0 MSun"] = dict(sse_age=2609167.84395 | units.yr,
                                      sse_mass=19.8895071502 | units.MSun,
                                      sse_rad=7.07713598137 | units.RSun,
                                      sse_L=51232.436257 | units.LSun,
                                      evtwin_age=2999999.99792 | units.yr,
                                      evtwin_mass=19.877338008 | units.MSun,
                                      evtwin_rad=6.77369109826 | units.RSun,
                                      evtwin_L=54123.7589452 | units.LSun)
        results["40.0 MSun"] = dict(sse_age=2627093.59096 | units.yr,
                                      sse_mass=38.3702723816 | units.MSun,
                                      sse_rad=13.6137662257 | units.RSun,
                                      sse_L=303498.877464 | units.LSun,
                                      evtwin_age=2999999.99923 | units.yr,
                                      evtwin_mass=37.0862227121 | units.MSun,
                                      evtwin_rad=13.5982316053 | units.RSun,
                                      evtwin_L=320099.608846 | units.LSun)

        for m in [40.] | units.MSun:  # ,20.,10.] | units.MSun:
            instance = FallbackStellarEvolution()
            instance._main_se.parameters.max_age_stop_condition = 3. | units.Myr

            star = Particle(1)
            star.mass = m
            star = instance.particles.add_particle(star)
            instance.commit_particles()

            while instance.ActiveModel[star].__class__.__name__ == "EVtwin":
                instance.evolve_model()

            print("%s\t%s\t%s\t%s\t%s\t%s" % (star.age, star.mass, star.radius,
             star.luminosity, star.stellar_type, instance.ActiveModel[star].__class__.__name__))

            self.assertAlmostRelativeEqual(results[str(m)]["sse_age"], instance._FBTimeseries[star].particles[0].SSEAgeAtSwitch, 7)
            self.assertAlmostRelativeEqual(results[str(m)]["sse_mass"], star.mass, 7)
            self.assertAlmostRelativeEqual(results[str(m)]["sse_rad"], star.radius, 7)
            self.assertAlmostRelativeEqual(results[str(m)]["sse_L"], star.luminosity, 7)

            star = instance._main_se.particles[0]

            self.assertAlmostRelativeEqual(results[str(m)]["evtwin_age"], star.age, 7)
            self.assertAlmostRelativeEqual(results[str(m)]["evtwin_mass"], star.mass, 2)
            self.assertAlmostRelativeEqual(results[str(m)]["evtwin_rad"], star.radius, 2)
            self.assertAlmostRelativeEqual(results[str(m)]["evtwin_L"], star.luminosity, 2)

            print("%s\t%s\t%s\t%s\t%s" % (star.age, star.mass, star.radius,
             star.luminosity, star.stellar_type))

            instance.stop()

    def slowtest3(self):
        print("Testing FallbackStellarEvolution: evolve 3 stars at the same time")


# results of original code  (not really check the numbers, tests have been relaxed because
# different timestepping of evolve_model and evolve_one_step)
        results = dict()
        results["10.0 MSun"] = dict(sse_age=3644655.52487 | units.yr,
                                      sse_mass=9.99289747724 | units.MSun,
                                      sse_rad=4.35949272485 | units.RSun,
                                      sse_L=6180.22077675 | units.LSun,
                                      evtwin_age=2999999.99732 | units.yr,
                                      evtwin_mass=9.99832250931 | units.MSun,
                                      evtwin_rad=4.1727904434 | units.RSun,
                                      evtwin_L=6175.5214329 | units.LSun)
        results["20.0 MSun"] = dict(sse_age=2609167.84395 | units.yr,
                                      sse_mass=19.8895071502 | units.MSun,
                                      sse_rad=7.30432858615 | units.RSun,
                                      sse_L=51232.436257 | units.LSun,
                                      evtwin_age=2999999.99792 | units.yr,
                                      evtwin_mass=19.877338008 | units.MSun,
                                      evtwin_rad=6.77369109826 | units.RSun,
                                      evtwin_L=54123.7589452 | units.LSun)
        results["40.0 MSun"] = dict(sse_age=2869735.64001 | units.yr,  # for some reason this is not 2617093.59096
                                      sse_mass=37.9172423183 | units.MSun,
                                      sse_rad=15.1388511451 | units.RSun,
                                      sse_L=322674.980598 | units.LSun,
                                      evtwin_age=2999999.99923 | units.yr,
                                      evtwin_mass=37.0862227121 | units.MSun,
                                      evtwin_rad=13.5982316053 | units.RSun,
                                      evtwin_L=320099.608846 | units.LSun)

        instance = FallbackStellarEvolution()
        instance._main_se.parameters.max_age_stop_condition = 3. | units.Myr

        stars = Particles(3, mass=[10., 20., 40.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        stars.initial_mass = stars.mass
        instance.commit_particles()

        while instance.model_time < .05 | units.Myr:
            instance.evolve_model()

        for star in stars:
            self.assertEqual(instance.ActiveModel[star].__class__.__name__, "EVtwin")

        while instance.model_time <= 3. | units.Myr:
            instance.evolve_model()

        for star in stars:
            self.assertEqual(instance.ActiveModel[star].__class__.__name__, "SSE")

        for star in stars:
            m = star.initial_mass
            self.assertAlmostRelativeEqual(results[str(m)]["sse_age"], instance._FBTimeseries[star].particles[0].SSEAgeAtSwitch, 3)

            evstar = star.as_particle_in_set(instance._main_se.particles)

            self.assertAlmostRelativeEqual(results[str(m)]["evtwin_age"], evstar.age, 7)
            self.assertAlmostRelativeEqual(results[str(m)]["evtwin_mass"], evstar.mass, 2)
            self.assertAlmostRelativeEqual(results[str(m)]["evtwin_rad"], evstar.radius, 2)
            self.assertAlmostRelativeEqual(results[str(m)]["evtwin_L"], evstar.luminosity, 2)

        instance.stop()

    def slowtest4(self):
        print("Testing FallbackStellarEvolution: evolve same 3 stars at the same time")

        instance = FallbackStellarEvolution()
        instance._main_se.parameters.max_age_stop_condition = 3. | units.Myr

        stars = Particles(3, mass=[40., 40., 40.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        instance.commit_particles()

        while instance.model_time <= 3. | units.Myr:
            instance.evolve_model()

        for star in stars:
            self.assertEqual(instance.ActiveModel[star].__class__.__name__, "SSE")

        age1 = instance._FBTimeseries[stars[0]].particles[0].SSEAgeAtSwitch
        age2 = instance._FBTimeseries[stars[1]].particles[0].SSEAgeAtSwitch
        age3 = instance._FBTimeseries[stars[2]].particles[0].SSEAgeAtSwitch

        self.assertEqual(age1, age2)
        self.assertEqual(age2, age3)

        instance.stop()

    def slowtest5(self):
        print("Testing FallbackStellarEvolution: evolve with end time")

        instance = FallbackStellarEvolution()
        instance._main_se.parameters.max_age_stop_condition = 0.1 | units.Myr

        stars = Particles(3, mass=[10., 20., 30.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        instance.commit_particles()

        instance.evolve_model(.101 | units.Myr)

        for star in stars:
            self.assertEqual(instance.ActiveModel[star].__class__.__name__, "SSE")

        self.assertTrue(.101 | units.Myr < instance.model_time)

        instance.stop()

    def test6(self):
        print("Testing FallbackStellarEvolution: enforce monotonic mass evolution")

        instance = FallbackStellarEvolution(enforce_monotonic_mass_evolution=True)
        instance._main_se.parameters.max_age_stop_condition = 3. | units.Myr

        stars = Particles(1, mass=[40.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        instance.commit_particles()

        while instance.ActiveModel[stars[0]].__class__.__name__ == "EVtwin":
            emass = stars[0].mass
            instance.evolve_model()

        self.assertTrue(stars[0].mass <= emass)

        instance.stop()

    def slowtest7(self):
        print("Testing FallbackStellarEvolution: test mesa")

        instance = FallbackStellarEvolution(MESA)
        instance._main_se.parameters.max_age_stop_condition = 3. | units.Myr

        stars = Particles(3, mass=[40., 40., 40.] | units.MSun)
        stars = instance.particles.add_particles(stars)
        instance.commit_particles()

        instance.evolve_model(3.01 | units.Myr)

        for star in stars:
            self.assertEqual(instance.ActiveModel[star].__class__.__name__, "SSE")

        age1 = instance._FBTimeseries[stars[0]].particles[0].SSEAgeAtSwitch
        age2 = instance._FBTimeseries[stars[1]].particles[0].SSEAgeAtSwitch
        age3 = instance._FBTimeseries[stars[2]].particles[0].SSEAgeAtSwitch

        self.assertEqual(age1, age2)
        self.assertEqual(age2, age3)

        instance.stop()
